"""
This file is part of gempy.

Created on 16.04.2019

@author: Elisa Heim
"""

import numpy as np
from scipy import fftpack
import pandas as pn
import os

try:
    from osgeo import gdal
    GDAL_IMPORT = True
except ImportError as e:
    GDAL_IMPORT = False
    print(e)

import matplotlib.pyplot as plt


class LoadDEMGDAL:
    """Class to include height elevation data (e.g. DEMs) with the geological grid
    """

    def __init__(self, path_dem, grid=None, extent=None, delete_temp=True):
        """
        Args:
            path_dem: path where dem is stored. file format: GDAL raster formats
            if grid: cropped to geomodel extent
        """
        if GDAL_IMPORT == False:
            raise ImportError('Gdal package is not installed. No support for raster formats.')
        self.dem = gdal.Open(path_dem)

        if isinstance(self.dem, type(None)):
            raise AttributeError('Raster file could not be opened {}. Check if the filepath is correct. If yes,'
                                 'check if your file fits the requirements of GDALs raster file formats.'.format(
                path_dem))

        try:
            self.dem_zval = self.dem.ReadAsArray()
        except AttributeError:
            raise AttributeError('Filepath seems to be wrong.')

        self._get_raster_dimensions()

        if extent is not None:
            self.regular_grid_extent = extent
            self.crop2grid()
        elif grid is not None:
            self.regular_grid_extent = grid.extent
            self.crop2grid()
        else:
            print('pass geo_model to automatically crop the DEM to the grid extent')
        print('depending on the size of the raster, this can take a while...')
        self.convert2xyz()

        if delete_temp is True:
            self.dem = None
            os.remove('topo.xyz')
            if os.path.exists('_cropped_DEM.tif'):
                os.remove('_cropped_DEM.tif')

    def _get_raster_dimensions(self):
        """calculates DEM extent, resolution, and max. z extent (d_z)"""
        ulx, xres, xskew, uly, yskew, yres = self.dem.GetGeoTransform()
        z = self.dem_zval
        if np.any(np.array([xskew, yskew])) != 0:
            print('DEM is not north-oriented.')
        lrx = ulx + (self.dem.RasterXSize * xres)
        lry = uly + (self.dem.RasterYSize * yres)
        self.resolution = np.array([(uly - lry) / (-yres), (lrx - ulx) / xres]).astype(int)
        self.extent = np.array([ulx, lrx, lry, uly]).astype(int)
        self.d_z = np.array([z.min(), z.max()])

    def get_values(self):
        return self.values_3D

    def info(self):
        ulx, xres, xskew, uly, yskew, yres = self.dem.GetGeoTransform()
        print('raster extent:  {}\n raster resolution: {}\n Pixel X size {}, Pixel Y size {}'.format(
            self.extent, self.resolution, xres, yres))
        plt.imshow(self.dem_zval, extent=self.extent)  # plot raster as image
        plt.colorbar()

    def crop2grid(self, delete_temp=True):
        """
        Crops raster to extent of the geomodel grid.
        """
        cornerpoints_geo = self._get_cornerpoints(self.regular_grid_extent)
        cornerpoints_dtm = self._get_cornerpoints(self.extent)

        # self.check()

        if np.any(cornerpoints_geo[:2] - cornerpoints_dtm[:2]) != 0:
            path_dest = '_cropped_DEM.tif'
            new_bounds = (self.regular_grid_extent[[0, 2, 1, 3]])
            gdal.Warp(path_dest, self.dem, options=gdal.WarpOptions(
                options=['outputBounds'], outputBounds=new_bounds))

            self.dem = gdal.Open(path_dest)
            self.dem_zval = self.dem.ReadAsArray()
            self._get_raster_dimensions()

        print('Cropped raster to geo_model.grid.extent.')

    def check(self):
        # TODO make this usable
        test = np.logical_and.reduce((self.regular_grid_extent[0] <= self.extent[0],
                                      self.regular_grid_extent[1] >= self.extent[1],
                                      self.regular_grid_extent[2] <= self.extent[2],
                                      self.regular_grid_extent[3] >= self.extent[3]))
        if test:
            cornerpoints_geo = self._get_cornerpoints(self.regular_grid_extent)
            cornerpoints_dtm = self._get_cornerpoints(self.extent)
            plt.scatter(cornerpoints_geo[:, 0], cornerpoints_geo[:, 1], label='grid extent')
            plt.scatter(cornerpoints_dtm[:, 0], cornerpoints_dtm[:, 1], label='raster extent')
            plt.legend(frameon=True, loc='upper left')
            raise AssertionError('The model extent is too different from the raster extent.')

    def convert2xyz(self, del_temp=True):
        """
        Translates the gdal raster object to a numpy array of xyz coordinates.
        """
        path_dest = 'topo.xyz'
        print('storing converted file...')
        shape = self.dem_zval.shape
        if len(shape) == 3:
            shape = shape[1:]
        gdal.Translate(path_dest, self.dem, options=gdal.TranslateOptions(options=['format'], format="XYZ"))

        xyz = pn.read_csv(path_dest, header=None, sep=' ').values
        self.values_3D = xyz.reshape((*shape, 3), order='C')

        # This is for 3D going from xyz to ijk
        self.values_3D = self.values_3D.swapaxes(0, 1)
        self.values_3D = np.flip(self.values_3D, 1)

        return self.values_3D

    def _resize(self, resx, resy):
        raise NotImplementedError

    def resample(self, new_xres, new_yres, save_path):
        """
        Decrease the pixel size of the raster.

        Args:
            new_xres (int): desired resolution in x-direction
            new_yres (int): desired resolution in y-direction
            save_path (str): filepath to where the output file should be stored

        Returns: Nothing, it writes a raster file with decreased resolution.

        """
        props = self.dem.GetGeoTransform()
        print('current pixel xsize:', props[1], 'current pixel ysize:', -props[-1])
        options = gdal.WarpOptions(options=['tr'], xRes=new_xres, yRes=new_yres)
        newfile = gdal.Warp(save_path, self.dem, options=options)
        newprops = newfile.GetGeoTransform()
        print('new pixel xsize:', newprops[1], 'new pixel ysize:', -newprops[-1])
        print('file saved in ' + save_path)

    def _get_cornerpoints(self, extent):
        """Get the coordinates of the bounding box.

        Args:
            extent: np.array([xmin, xmax, ymin, ymax)]

        Returns: np.ndarray with corner coordinates

        """
        upleft = ([extent[0], extent[3]])
        lowleft = ([extent[0], extent[2]])
        upright = ([extent[1], extent[3]])
        lowright = ([extent[1], extent[2]])
        return np.array([upleft, lowleft, upright, lowright])


class LoadDEMArtificial:

    def __init__(self, grid=None, fd=2.0, extent=None, resolution=None, d_z=None):
        """Class to create a random topography based on a fractal grid algorithm.

        Args:
            fd:         fractal dimension, defaults to 2.0
            d_z:        maximum height difference. If none, last 20% of the model in z direction
            extent:     extent in xy direction. If none, geo_model.grid.extent
            resolution: desired resolution of the topography array. If none, geo_model.grid.resolution
        """
        self.values_2d = np.array([])

        self.resolution = grid.resolution[:2] if resolution is None else resolution

        assert all(np.asarray(self.resolution) >= 2), 'The regular grid needs to be at least of size 2 on all ' \
                                                      'directions.'
        self.extent = grid.extent if extent is None else extent

        if d_z is None:
            self.d_z = np.array(
                [self.extent[5] - (self.extent[5] - self.extent[4]) * 1 / 5,
                 self.extent[5]])
            print(self.d_z)
        else:
            self.d_z = d_z

        topo = self.fractalGrid(fd, n=self.resolution.max())
        topo = np.interp(topo, (topo.min(), topo.max()), self.d_z)

        self.dem_zval = topo[:self.resolution[0], :self.resolution[1]]  # crop fractal grid with resolution
        self.create_topo_array()

    def fractalGrid(self, fd, n=256):
        """
        Modified after https://github.com/samthiele/pycompass/blob/master/examples/3_Synthetic%20Examples.ipynb

        Generate isotropic fractal surface image using
        spectral synthesis method [1, p.]
        References:
        1. Yuval Fisher, Michael McGuire,
        The Science of Fractal Images, 1988

        (cf. http://shortrecipes.blogspot.com.au/2008/11/python-isotropic-fractal-surface.html)
        **Arguments**:
         -fd = the fractal dimension
         -N = the size of the fractal surface/image

        """
        h = 1 - (fd - 2)
        # X = np.zeros((N, N), complex)
        a = np.zeros((n, n), complex)
        powerr = -(h + 1.0) / 2.0

        for i in range(int(n / 2) + 1):
            for j in range(int(n / 2) + 1):
                phase = 2 * np.pi * np.random.rand()

                if i != 0 or j != 0:
                    rad = (i * i + j * j) ** powerr * np.random.normal()
                else:
                    rad = 0.0

                a[i, j] = complex(rad * np.cos(phase), rad * np.sin(phase))

                if i == 0:
                    i0 = 0
                else:
                    i0 = n - i

                if j == 0:
                    j0 = 0
                else:
                    j0 = n - j

                a[i0, j0] = complex(rad * np.cos(phase), -rad * np.sin(phase))

                a.imag[int(n / 2)][0] = 0.0
                a.imag[0, int(n / 2)] = 0.0
                a.imag[int(n / 2)][int(n / 2)] = 0.0

        for i in range(1, int(n / 2)):
            for j in range(1, int(n / 2)):
                phase = 2 * np.pi * np.random.rand()
                rad = (i * i + j * j) ** powerr * np.random.normal()
                a[i, n - j] = complex(rad * np.cos(phase), rad * np.sin(phase))
                a[n - i, j] = complex(rad * np.cos(phase), -rad * np.sin(phase))

        itemp = fftpack.ifft2(a)
        itemp = itemp - itemp.min()

        return itemp.real / itemp.real.max()

    def create_topo_array(self):
        """for masking the lith block"""
        x = np.linspace(self.extent[0], self.extent[1], self.resolution[0])
        y = np.linspace(self.extent[2], self.extent[3], self.resolution[1])
        self.x = x
        self.y = y
        xx, yy = np.meshgrid(x, y, indexing='ij')
        self.values_2d = np.dstack([xx, yy, self.dem_zval])

    def get_values(self):
        return self.values_2d



import gempy_legacy as gp
import pandas as pn
import numpy as np
import os
import pytest
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
input_path = os.path.dirname(__file__)+'/../../notebooks/data'


class TestFabianModel:

    def test_init_model(self, one_fault_model):
        print(one_fault_model)

    def test_get_data(self, one_fault_model):
        print(gp.get_data(one_fault_model))
        print(gp.get_data(one_fault_model, itype='additional_data'))

    def test_plotting_data(self, one_fault_model):
        gp.plot.plot_2d(one_fault_model, show_data=True, show_results=False)

    def test_compute_model(self, one_fault_model_solution):
        sol = one_fault_model_solution.solutions
        print(sol)
        return sol

    def test_plot_section(self, one_fault_model):

        gp.plot.plot_2d(one_fault_model, cell_number=25,
                        direction='y', show_data=True)

       # plt.savefig(os.pardir+'/../figs/example1.png')




import gempy_legacy as gp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import pandas as pn
import numpy as np
import os
import pytest
input_path = os.path.dirname(__file__)+'/../input_data'

# ## Preparing the Python environment
#
# For modeling with GemPy, we first need to import it. We should also import any other packages we want to
# utilize in our Python environment.Typically, we will also require `NumPy` and `Matplotlib` when working
# with GemPy. At this point, we can further customize some settings as desired, e.g. the size of figures or,
# as we do here, the way that `Matplotlib` figures are displayed in our notebook (`%matplotlib inline`).


# These two lines are necessary only if GemPy is not installed
import sys, os
sys.path.append("../..")


@pytest.fixture(scope="module")
def load_model():
    verbose = False
    geo_model = gp.create_model('Model_Tuto1-1')

    # Importing the data from CSV-files and setting extent and resolution
    gp.init_data(geo_model, [0, 2000., 0, 2000., 0, 2000.], [50 ,50 ,50],
                 path_o=input_path+"/simple_fault_model_orientations.csv",
                 path_i=input_path+"/simple_fault_model_points.csv", default_values=True)

    df_cmp_i = gp.get_data(geo_model, 'surface_points')
    df_cmp_o = gp.get_data(geo_model, 'orientations')

    df_o = pn.read_csv(input_path + "/simple_fault_model_orientations.csv")
    df_i = pn.read_csv(input_path + "/simple_fault_model_points.csv")

    assert not df_cmp_i.empty, 'data was not set to dataframe'
    assert not df_cmp_o.empty, 'data was not set to dataframe'
    assert df_cmp_i.shape[0] == df_i.shape[0], 'data was not set to dataframe'
    assert df_cmp_o.shape[0] == df_o.shape[0], 'data was not set to dataframe'

    if verbose:
        gp.get_data(geo_model, 'surface_points').head()

    return geo_model


def test_load_model_df():

    verbose = True
    df_i = pn.DataFrame(np.random.randn(6,3), columns='X Y Z'.split())
    df_i['formation'] = ['surface_1' for _ in range(3)] + ['surface_2' for _ in range(3)]

    df_o = pn.DataFrame(np.random.randn(6,6), columns='X Y Z azimuth dip polarity'.split())
    df_o['formation'] = ['surface_1' for _ in range(3)] + ['surface_2' for _ in range(3)]

    geo_model = gp.create_model('test')
    # Importing the data directly from the dataframes
    gp.init_data(geo_model, [0, 2000., 0, 2000., 0, 2000.], [50, 50, 50],
                 surface_points_df=df_i, orientations_df=df_o, default_values=True)

    df_cmp_i = gp.get_data(geo_model, 'surface_points')
    df_cmp_o = gp.get_data(geo_model, 'orientations')

    if verbose:
        print(df_cmp_i.head())
        print(df_cmp_o.head())

    assert not df_cmp_i.empty, 'data was not set to dataframe'
    assert not df_cmp_o.empty, 'data was not set to dataframe'
    assert df_cmp_i.shape[0] == 6, 'data was not set to dataframe'
    assert df_cmp_o.shape[0] == 6, 'data was not set to dataframe'

    # try without the default_values command

    geo_model = gp.create_model('test')
    # Importing the data directly from the dataframes
    gp.init_data(geo_model, [0, 2000., 0, 2000., 0, 2000.], [50 ,50 ,50],
                 surface_points_df=df_i, orientations_df=df_o)

    df_cmp_i2 = gp.get_data(geo_model, 'surface_points')
    df_cmp_o2 = gp.get_data(geo_model, 'orientations')

    if verbose:
        print(df_cmp_i2.head())
        print(df_cmp_o2.head())

    assert not df_cmp_i2.empty, 'data was not set to dataframe'
    assert not df_cmp_o2.empty, 'data was not set to dataframe'
    assert df_cmp_i2.shape[0] == 6, 'data was not set to dataframe'
    assert df_cmp_o2.shape[0] == 6, 'data was not set to dataframe'

    return geo_model


@pytest.fixture(scope='module')
def map_sequential_pile(load_model):
    geo_model = load_model

    # TODO decide what I do with the layer order

    gp.map_stack_to_surfaces(geo_model, {"Fault_Series": 'Main_Fault',
                                          "Strat_Series": ('Sandstone_2', 'Siltstone',
                                                           'Shale', 'Sandstone_1', 'basement')},
                             remove_unused_series=True)

    geo_model.set_is_fault(['Fault_Series'])
    return geo_model


def test_get_data(load_model):
    geo_model = load_model
    return gp.get_data(geo_model, 'orientations').head()


def test_define_sequential_pile(map_sequential_pile):
    print(map_sequential_pile._surfaces)


def test_sequential_pile_colors(load_model):
    geo_model = load_model

    gp.map_stack_to_surfaces(geo_model, {"Fault_Series": 'Main_Fault',
                                          "Strat_Series": ('Sandstone_2', 'Siltstone',
                                                           'Shale', 'Sandstone_1', 'basement')},
                             remove_unused_series=True)

    color1 = geo_model._surfaces.colors.colordict['Main_Fault']
    geo_model.set_is_fault(['Fault_Series'], toggle=True)
    color2 = geo_model._surfaces.colors.colordict['Main_Fault']
    geo_model.set_is_fault(['Fault_Series'], toggle=True)
    color3 = geo_model._surfaces.colors.colordict['Main_Fault']
    assert color1 == color3
    #print(color1, color2, color3)


def test_compute_model(interpolator, map_sequential_pile):
    geo_model = map_sequential_pile
    geo_model.set_aesara_graph(interpolator)

    gp.compute_model(geo_model, compute_mesh=False)

    test_values = [45, 150, 2500]
    if False:
        np.save(input_path+'/test_integration_lith_block.npy', geo_model.solutions.lith_block[test_values])

    # Load model
    real_sol = np.load(input_path + '/test_integration_lith_block.npy')

    # We only compare the block because the absolute pot field I changed it
    np.testing.assert_array_almost_equal(np.round(geo_model.solutions.lith_block[test_values]), real_sol, decimal=0)

    gp.plot.plot_2d(geo_model, cell_number=25,
                    direction='y', show_data=True)
    plt.savefig(os.path.dirname(__file__)+'/../figs/test_integration_lith_block')

    gp.plot.plot_2d(geo_model, cell_number=25, series_n=1, N=15, show_scalar=True,
                    direction='y', show_data=True)

    plt.savefig(os.path.dirname(__file__)+'/../figs/test_integration_scalar')


def test_save_model(interpolator, map_sequential_pile):
    geo_model = map_sequential_pile
    geo_model.set_aesara_function(interpolator)
    gp.compute_model(geo_model, compute_mesh=False)
    gp.save_model(geo_model, name='test_save_model', path=input_path+'/save_model/', save_solution=True,
                  compress=False)

    lith_block_sol = np.load(input_path+'/save_model/test_save_model/test_save_model_lith_block.npy')
    np.testing.assert_array_almost_equal(geo_model.solutions.lith_block, lith_block_sol, decimal=0)


def test_kriging_mutation(interpolator, map_sequential_pile):
    geo_model = map_sequential_pile
    geo_model.set_aesara_graph(interpolator)

    gp.compute_model(geo_model, compute_mesh=False)
    gp.plot.plot_2d(geo_model, cell_number=25, show_scalar=True, series_n=1, N=15,
                    direction='y', show_data=True)
    print(geo_model.solutions.lith_block, geo_model._additional_data)
    #plt.savefig(os.path.dirname(__file__)+'/figs/test_kriging_mutation')

    geo_model.modify_kriging_parameters('range', 1)
    geo_model.modify_kriging_parameters('drift equations', [0, 3])

    print(geo_model.solutions.lith_block, geo_model._additional_data)
    # copy dataframe before interpolator is calculated
    pre = geo_model._additional_data.kriging_data.df.copy()

    gp.set_interpolator(geo_model, compile_aesara=True,
                        aesara_optimizer='fast_compile', update_kriging=False)
    gp.compute_model(geo_model, compute_mesh=False)

    gp.plot.plot_2d(geo_model, cell_number=25, series_n=1, N=15, show_boundaries=False,
                    direction='y', show_data=True, show_lith=True)

    print(geo_model.solutions.lith_block, geo_model._additional_data)
    plt.savefig(os.path.dirname(__file__)+'/../figs/test_kriging_mutation2')
    assert geo_model._additional_data.kriging_data.df['range'][0] == pre['range'][0]


"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.

@author: Alexander Schaaf, Miguel de la Varga
"""

import warnings
try:
    import pymc
except ImportError:
    warnings.warn("pymc (v2) package is not installed. No support for stochastic simulation posterior analysis.")
try:
    import pymc
except ImportError:
    warnings.warn("pymc (v3) package is not installed. No support for stochastic simulation posterior analysis.")
import numpy as np
import pandas as pn
import gempy_legacy as gp
try:
    import tqdm
except ImportError:
    warnings.warn("tqdm package not installed. No support for dynamic progress bars.")


def change_input_data_general(data_array, interp_data):
    """
    Changes input data in interp_data to posterior input data at iteration i.

    Args:
        data_array (np.ndarray):
        interp_data (gempy.data_management.InterpolationData): An interp_data object with the structure we want to
        compute.

    Returns:
         None, in-place operation
    """
    # get mask for interfaces and orientations by finding nan values
    interf_mask = np.isnan(data_array[:, -1])
    orient_mask = np.invert(interf_mask)

    # replace interface data
    interp_data.geo_data_res.interfaces[["X", "Y", "Z"]] = data_array[interf_mask, :3]
    # replace foliation data
    if data_array[orient_mask, :].shape[1] == 6:
        interp_data.geo_data_res._orientations[["X", "Y", "Z", "dip", "azimuth", "polarity"]] = data_array[orient_mask, :]
    elif data_array[orient_mask, :].shape[1] == 9:
        interp_data.geo_data_res._orientations[["X", "Y", "Z", "G_x", "G_y", "G_z", "dip", "azimuth", "polarity"]] = data_array[orient_mask, :]
    else:
        raise ValueError("Value mismatch.")
    # recalc gradients
    recalc_gradients(interp_data.geo_data_res._orientations)
    # update interpolator
    interp_data.update_interpolator()

    return None


def change_input_data(db, interp_data, i):
    """
    Changes input data in interp_data to posterior input data at iteration i.

    Args:
        interp_data (gempy.data_management.InterpolationData): An interp_data object with the structure we want to
        compute.
        i (int): Iteration we want to recompute

    Returns:
         gempy.data_management.InterpolationData: interp_data with the data of the given iteration
    """
    i = int(i)
    # replace interface data
    interp_data.geo_data_res._surface_points[["X", "Y", "Z"]] = db.trace("input_interf")[i]
    # replace foliation data
    try:
        interp_data.geo_data_res._orientations[["X", "Y", "Z", "dip", "azimuth", "polarity"]] = db.trace("input_orient")[i]
    except ValueError:
        interp_data.geo_data_res._orientations[["G_x", "G_y", "G_z", "X", "Y", "Z", "dip", "azimuth", "polarity"]] = db.trace("input_orient")[i]

    recalc_gradients(interp_data.geo_data_res._orientations)

    # interp_data.geo_data_res_no_basement = interp_data.rescale_data(interp_data.geo_data_res_no_basement)

    # update interpolator
    interp_data.update_interpolator()
    if verbose:
        print("interp_data parameters changed.")
    return interp_data


def change_input_data_avrg(db, interp_data):
    """Calculates average posterior parameters and updates interp_data.geo_data_res dataframes."""

    # get input data
    trace = db.geo_model.gettrace()
    # set average in categories_df
    interp_data.geo_data_res._surface_points[["X", "Y", "Z"]] = trace[:, 0][:].mean(axis=0)
    interp_data.geo_data_res._orientations[["X", "Y", "Z", "dip", "azimuth", "polarity"]] = trace[:, 1][:].mean(axis=0)
    recalc_gradients(interp_data.geo_data_res._orientations)  # recalc gradients
    interp_data.update_interpolator()  # update interpolator
    return interp_data


def _change_input_data_old(db, interp_data, i, tracename="input_data"):
    """
    Changes input data in interp_data to posterior input data at iteration i.

    Args:
        interp_data (gempy.data_management.InterpolationData): An interp_data object with the structure we want to
        compute.
        i (int): Iteration we want to recompute

    Returns:
         gempy.data_management.InterpolationData: interp_data with the data of the given iteration
    """
    i = int(i)
    # replace interface data
    interp_data.geo_data_res._surface_points[["X", "Y", "Z"]] = db.trace(tracename)[i][0]
    # replace foliation data
    try:
        interp_data.geo_data_res._orientations[["X", "Y", "Z", "dip", "azimuth", "polarity"]] = db.trace(tracename)[i][1]
    except ValueError:
        interp_data.geo_data_res._orientations[["G_x", "G_y", "G_z", "X", "Y", "Z", "dip", "azimuth", "polarity"]] = db.trace(tracename)[i][1]

    recalc_gradients(interp_data.geo_data_res._orientations)

    # interp_data.geo_data_res_no_basement = interp_data.rescale_data(interp_data.geo_data_res_no_basement)

    # update interpolator
    interp_data.update_interpolator()

    return interp_data


def compute_posterior_models_all(db, interp_data, indices, u_grade=None, get_potential_at_surface_points=False):
    """Computes block models from stored input parameters for all iterations.

    Args:
        db (): loaded pymc database (e.g. hdf5)
        interp_data  (gp.data_management.InterpolatorData): GemPy interpolator object
        indices (list or np.array): Trace indices specifying which models from the database will be calculated.
        u_grade (list, optional):
        get_potential_at_surface_points:

    Returns:

    """

    for i in tqdm.tqdm(indices):
        interp_data_loop = change_input_data(db, interp_data, i)
        lb, fb = gp.compute_model(interp_data_loop, output="geology", u_grade=u_grade, get_potential_at_surface_points=get_potential_at_surface_points)
        if i == 0 or i == indices[0]:
            lbs = np.expand_dims(lb, 0)
            fbs = np.expand_dims(fb, 0)
        else:
            lbs = np.concatenate((lbs, np.expand_dims(lb, 0)), axis=0)
            fbs = np.concatenate((fbs, np.expand_dims(fb, 0)), axis=0)

    return lbs, fbs


def calculate_probability_lithology(lith_blocks):
    """Blocks must be just the lith blocks!"""
    lith_id = np.unique(lith_blocks)
    # lith_count = np.zeros_like(lith_blocks[0:len(lith_id)])
    lith_count = np.zeros((len(np.unique(lith_blocks)), lith_blocks.shape[1]))
    for i, l_id in enumerate(lith_id):
        lith_count[i] = np.sum(lith_blocks == l_id, axis=0)
    lith_prob = lith_count / len(lith_blocks)
    return lith_prob


def calculate_information_entropy(lith_prob):
    """Calculates information entropy for the given probability array."""
    ie = np.zeros_like(lith_prob[0])
    for l in lith_prob:
        pm = np.ma.masked_equal(l, 0)  # mask where layer prob is 0
        ie -= (pm * np.ma.log2(pm)).filled(0)
    return ie


def calculate_information_entropy_total(ie, absolute=False):
    """Calculate total information entropy (float) from an information entropy array."""
    if absolute:
        return np.sum(ie)
    else:
        return np.sum(ie) / np.size(ie)


def calculate_ie(lbs):
    """Computes the per-voxel and total information entropy of given block models."""

    lith_prob = calculate_probability_lithology(lbs)
    ie = calculate_information_entropy(lith_prob)
    ie_total = calculate_information_entropy_total(ie)

    return ie, ie_total


class Posterior:
    def __init__(self, dbname, pymc_model_f="gempy_model", pymc_topo_f="gempy_topo",
                 topology=False, verbose=False):
        """
        Posterior database analysis for GemPy-pymc2 hdf5 databases.
        Args:
            dbname (str): Path of the hdf5 database.
            pymc_model_f (str, optional): name of the model output function used (default: "gempy_model).
            pymc_topo_f (str, optional): name of the topology output function used (default: "gempy_topo).
            topology (bool, optional):  if a topology trace should be loaded from the database (default: False).
            verbose (bool, optional): Verbosity switch.
        """
        # TODO: Add a method to set the lith_block and fault_block


        self.verbose = verbose
        # load db
        self.db = pymc.database.hdf5.load(dbname)

        self.n_iter = self.db.getstate()['sampler']['_iter'] - self.db.getstate()["sampler"]["_burn"]
        # get trace names
        self.trace_names = self.db.trace_names[0]

        # TODO DEP
        # get gempy block models
        # try:
        #     self.lb = self.db.trace(pymc_model_f)[:, :2, :]
        #     self.fb = self.db.trace(pymc_model_f)[:, 2:, :]
        # except KeyError:
        #     print("No GemPy model trace tallied.")
        #     self.lb = None
        #     self.fb = None

        if topology:  # load topology data from database
            topo_trace = self.db.trace(pymc_topo_f)[:]
            # load graphs
            self.topo_graphs = topo_trace[:, 0]
            # load centroids
            self.topo_centroids = topo_trace[:, 1]
            # unique labels
            self.topo_labels_unique = topo_trace[:, 2]
            # get the look-up-tables
            self.topo_lith_to_labels_lot = topo_trace[:, 3]
            self.topo_labels_to_lith_lot = topo_trace[:, 4]
            del topo_trace

            self.topo_unique, self.topo_unique_freq, self.topo_unique_ids, self.topo_unique_prob = (None, None, None, None)
            self.topo_count_dict = None

            self.topo_analyze()

        # load input data
        self.input_data = self.db.geo_model.gettrace()

        self.lith_prob = None
        self.ie = None
        self.ie_total = None

    def change_input_data(self, interp_data, i):
        """
        Changes input data in interp_data to posterior input data at iteration i.

        Args:
            interp_data (gempy.data_management.InterpolationData): An interp_data object with the structure we want to
            compute.
            i (int): Iteration we want to recompute

        Returns:
             gempy.data_management.InterpolationData: interp_data with the data of the given iteration
        """
        i = int(i)
        # replace interface data
        interp_data.geo_data_res._surface_points[["X", "Y", "Z"]] = self.input_data[i][0]
        # replace foliation data
        interp_data.geo_data_res._orientations[["G_x", "G_y", "G_z", "X", "Y", "Z", "dip", "azimuth", "polarity"]] = self.input_data[i][1]

        recalc_gradients(interp_data.geo_data_res._orientations)

        # update interpolator
        interp_data.update_interpolator()
        if self.verbose:
            print("interp_data parameters changed.")
        return interp_data

    def change_input_data_avrg(self, interp_data):
        """Calculates average posterior parameters and updates interp_data.geo_data_res dataframes."""

        # average input data
        interf_avrg = self.input_data[:, 0][:].mean(axis=0)
        orient_avrg = self.input_data[:, 1][:].mean(axis=0)
        # set average in categories_df
        interp_data.geo_data_res._surface_points[["X", "Y", "Z"]] = interf_avrg
        interp_data.geo_data_res._orientations[["G_x", "G_y", "G_z", "X", "Y", "Z", "dip", "azimuth", "polarity"]] = orient_avrg
        recalc_gradients(interp_data.geo_data_res._orientations)  # recalc gradients
        interp_data.update_interpolator()  # update interpolator
        return interp_data

    def compute_entropy(self):
        """Computes the voxel information entropy of stored block models."""
        if self.lb is None:
            return "No models stored in self.lb, please run 'self.compute_posterior_models_all' to generate block" \
                   " models for all iterations."

        self.lith_prob = calculate_probability_lithology(self.lb[:, 0, :])
        self.ie = calculate_information_entropy(self.lith_prob)
        self.ie_total = calculate_information_entropy_total(self.ie)
        print("Information Entropy successfully calculated. Stored in self.ie and self.ie_total")

    def topo_count_connection(self, n1, n2):
        """Counts the amount of times connection between nodes n1 and n2 in all of the topology graphs."""
        count = 0
        for G in self.topo_graphs:
            count += gp.topology.check_adjacency(G, n1, n2)
        return count

    def topo_count_connection_array(self, n1, n2):
        count = []
        for G in self.topo_graphs:
            count.append(gp.topology.check_adjacency(G, n1, n2))
        return count

    def topo_count_total_number_of_nodes(self):
        """Counts the amount of topology graphs with a certain amount of total nodes."""
        self.topo_count_dict = {}
        for g in self.topo_graphs:
            c = len(g.adj.keys())
            if c in self.topo_count_dict.keys():
                self.topo_count_dict[c] += 1
            else:
                self.topo_count_dict[c] = 1

    def topo_analyze(self):
        """Analysis of the tallied topology distribution."""
        if self.verbose:
            print("Starting topology analysis. This could take a while (depending on # iterations).")
        self.topo_unique, self.topo_unique_freq, self.topo_unique_ids = get_unique_topo(self.topo_graphs)
        self.topo_unique_prob = self.topo_unique_freq / np.sum(self.topo_unique_freq)
        # count unique node numbers
        self.topo_count_total_number_of_nodes()

        self.topo_sort = np.argsort(self.topo_unique_freq)[::-1]

        if self.verbose:
            print("Topology analysis completed.")


class PosteriorPyMC3(Posterior):
    def __init__(self, dbname, pymc_model_f="gempy_model", pymc_topo_f="gempy_topo",
                 topology=False, verbose=False):
        """
        Posterior database analysis for GemPy-PyMC3 hdf5 databases.
        Args:
            dbname (str): Path of the hdf5 database.
            pymc_model_f (str, optional): name of the model output function used (default: "gempy_model).
            pymc_topo_f (str, optional): name of the topology output function used (default: "gempy_topo).
            topology (bool, optional):  if a topology trace should be loaded from the database (default: False).
            verbose (bool, optional): Verbosity switch.
        """
        # TODO: Add a method to set the lith_block and fault_block


        self.verbose = verbose
        # load db
        with pymc3.Model() as model:
            self.db = pymc3.backends.hdf5.load(dbname)
            # model environment required? / alternatives?

        self.n_iter = self.db.get_values(pymc_model_f).shape[0]-1
        # get trace names
        self.trace_names = self.db.varnames

        # TODO DEP
        # get gempy block models
        # try:
        #     self.lb = self.db.trace(pymc_model_f)[:, :2, :]
        #     self.fb = self.db.trace(pymc_model_f)[:, 2:, :]
        # except KeyError:
        #     print("No GemPy model trace tallied.")
        #     self.lb = None
        #     self.fb = None

        if topology:  # load topology data from database
            topo_trace = self.db.trace(pymc_topo_f)[:]
            # load graphs
            self.topo_graphs = topo_trace[:, 0]
            # load centroids
            self.topo_centroids = topo_trace[:, 1]
            # unique labels
            self.topo_labels_unique = topo_trace[:, 2]
            # get the look-up-tables
            self.topo_lith_to_labels_lot = topo_trace[:, 3]
            self.topo_labels_to_lith_lot = topo_trace[:, 4]
            del topo_trace

            self.topo_unique, self.topo_unique_freq, self.topo_unique_ids, self.topo_unique_prob = (None, None, None, None)
            self.topo_count_dict = None

            self.topo_analyze()

        # load input data
        #self.input_data = self.db.input_data.gettrace()

        self.lith_prob = None
        self.ie = None
        self.ie_total = None

    def change_input_data(self, interp_data, i):
        """
        Changes input data in interp_data to posterior input data at iteration i.

        Args:
            interp_data (gempy.data_management.InterpolationData): An interp_data object with the structure we want to
            compute.
            i (int): Iteration we want to recompute

        Returns:
             gempy.data_management.InterpolationData: interp_data with the data of the given iteration
        """
        i = int(i)
        # replace res and ref point data
        interp_data.get_input_data()[4] = self.db.get_values('input_ref')[i]
        interp_data.get_input_data()[5] = self.db.get_values('input_res')[i]

        #recalc_gradients(interp_data.geo_data_res.orientations)

        # update interpolator
        interp_data.update_interpolator()
        if self.verbose:
            print("interp_data parameters changed.")
        return interp_data



def find_first_match(t, topo_u):
    index = 0
    for t2 in topo_u:
        if gp.topology.compare_graphs(t, t2) == 1:
            return index  # the models match
        index += 1

    return -1


def get_unique_topo(topo_l):
    # create list for our unique topologies
    topo_u = []
    topo_u_freq = []
    topo_u_ids = np.empty_like(topo_l)

    for n, t in enumerate(topo_l):
        i = find_first_match(t, topo_u)
        if i == -1:  # is a yet unobserved topology, so append it and initiate frequency
            topo_u.append(t)
            topo_u_freq.append(1)
            topo_u_ids[n] = len(topo_u) - 1
        else:  # is a known topology
            topo_u_freq[i] += 1  # 1-up the corresponding frequency
            topo_u_ids[n] = i

    return topo_u, topo_u_freq, topo_u_ids


def get_unique_jaccard(js):
    j_u = np.unique(js)  # unique topology states


def modify_plane_dip(dip, group_id, data_obj):
    """Modify a dip angle of a plane identified by a group_id, recalculate the gradient and move the points vertically.
    Currently only supports the modification of dip angle - azimuth and polarity will stay the same.

    Args:
        dip (float): Desired dip angle of the plane.
        group_id (str): Group id identifying the data points belonging to the plane.
        data_obj (:obj:): Data object to be modified (geo_model or interp_data.geo_data_res_no_basement)

    Returns:
        Directly modifies the given data object.
    """
    # get foliation and interface data points ids
    fol_f = data_obj._orientations["group_id"] == group_id
    interf_f = data_obj._surface_points["group_id"] == group_id

    # get indices
    interf_i = data_obj._surface_points[interf_f].index
    fol_i = data_obj._orientations[fol_f].index[0]

    # update dip value for orientations
    data_obj._orientations.set_value(fol_i, "dip", dip)
    # get azimuth and polarity
    az = float(data_obj._orientations.iloc[fol_i]["azimuth"])
    pol = data_obj._orientations.iloc[fol_i]["polarity"]

    # calculate gradient/normal and modify
    gx, gy, gz = calculate_gradient(dip, az, pol)
    data_obj._orientations.set_value(fol_i, "G_x", gx)
    data_obj._orientations.set_value(fol_i, "G_y", gy)
    data_obj._orientations.set_value(fol_i, "G_z", gz)

    normal = [gx, gy, gz]
    centroid = np.array([float(data_obj._orientations[fol_f]["X"]),
                         float(data_obj._orientations[fol_f]["Y"]),
                         float(data_obj._orientations[fol_f]["Z"])])
    # move points vertically to fit plane
    move_plane_points(normal, centroid, data_obj, interf_f)


def move_plane_points(normal, centroid, data_obj, interf_f):
    """Moves interface points to fit plane of given normal and centroid in data object."""
    a, b, c = normal
    d = -a * centroid[0] - b * centroid[1] - c * centroid[2]
    for i, row in data_obj._surface_points[interf_f].iterrows():
        # iterate over each point and recalculate Z, set Z
        Z = (a * row["X"] + b * row["Y"] + d) / -c
        data_obj._surface_points.set_value(i, "Z", Z)


def calculate_gradient(dip, az, pol):
    """Calculates the gradient from dip, azimuth and polarity values."""
    g_x = np.sin(np.deg2rad(dip)) * np.sin(np.deg2rad(az)) * pol
    g_y = np.sin(np.deg2rad(dip)) * np.cos(np.deg2rad(az)) * pol
    g_z = np.cos(np.deg2rad(dip)) * pol
    return g_x, g_y, g_z


def recalc_gradients(folations_dataframe):
    folations_dataframe["G_x"] = np.sin(np.deg2rad(folations_dataframe["dip"].astype('float'))) * \
                             np.sin(np.deg2rad(folations_dataframe["azimuth"].astype('float'))) * \
                             folations_dataframe["polarity"].astype('float')
    folations_dataframe["G_y"] = np.sin(np.deg2rad(folations_dataframe["dip"].astype('float'))) * \
                             np.cos(np.deg2rad(folations_dataframe["azimuth"].astype('float'))) *\
                             folations_dataframe["polarity"].astype('float')
    folations_dataframe["G_z"] = np.cos(np.deg2rad(folations_dataframe["dip"].astype('float'))) *\
                             folations_dataframe["polarity"].astype('float')



import gempy_legacy as gp
import numpy as np
import os
input_path = os.path.dirname(__file__) + '/../../examples/data'


def test_set_orientations():
    # Importing the data from CSV-files and setting extent and resolution
    geo_data = gp.create_data(extent=[0, 2000, 0, 2000, 0, 2000], resolution=[50, 50, 50],
                              path_o=input_path + '/input_data/tut_chapter1/simple_fault_model_orientations.csv',
                              path_i=input_path + '/input_data/tut_chapter1/simple_fault_model_points.csv')

    gp.get_data(geo_data)

    # Assigning series to formations as well as their order (timewise)
    gp.map_stack_to_surfaces(geo_data, {"Fault_Series": 'Main_Fault',
                                         "Strat_Series": ('Sandstone_2', 'Siltstone')})

    geo_data._orientations.create_orientation_from_surface_points(geo_data.surface_points, [0, 1, 2])

    gp.set_orientation_from_surface_points(geo_data, [0, 1, 2])
    

def test_select_nearest_surface_points():
    data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
    path_to_data = data_path + "/data/input_data/jan_models/"

    geo_data = gp.create_data('fault', extent=[0, 1000, 0, 1000, 0, 1000],
                              resolution=[50, 50, 50],
                              path_o=path_to_data + "model5_orientations.csv",
                              path_i=path_to_data + "model5_surface_points.csv")

    # Assigning series to formations as well as their order (timewise)
    gp.map_stack_to_surfaces(geo_data, {"Fault_Series": 'fault',
                                        "Strat_Series": ('rock2', 'rock1')})
    geo_data.set_is_fault(['Fault_Series'])

    # detect fault names
    f_id = geo_data._faults.df.index.categories[
        geo_data._faults.df.isFault.values]
    # find fault points
    fault_poi = geo_data._surface_points.df[
        geo_data._surface_points.df.series.isin(f_id)]

    # find neighbours
    knn = gp.select_nearest_surfaces_points(geo_data, fault_poi, 1)
    radius = gp.select_nearest_surfaces_points(geo_data, fault_poi, 200.)

    # sort neighbours, necessary for equal testing
    knn = [k.sort_values() for k in knn]
    radius = [r.sort_values() for r in radius]

    # define reference
    reference = [[16,17],[16,17],[18,19],[18,19],[20,21],[20,21]]

    assert np.array_equal(reference, knn) and np.array_equal(reference, radius)


def test_set_orientation_from_neighbours():
    data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
    path_to_data = data_path + "/data/input_data/jan_models/"

    geo_data = gp.create_data('fault', extent=[0, 1000, 0, 1000, 0, 1000],
                              resolution=[50, 50, 50],
                          path_o=path_to_data + "model5_orientations.csv",
                          path_i=path_to_data + "model5_surface_points.csv")

    # Assigning series to formations as well as their order (timewise)
    gp.map_stack_to_surfaces(geo_data, {"Fault_Series": 'fault',
                                    "Strat_Series": ('rock2', 'rock1')})
    geo_data.set_is_fault(['Fault_Series'])

    # detect fault names
    f_id = geo_data._faults.df.index.categories[
        geo_data._faults.df.isFault.values]
    # find fault points
    fault_poi = geo_data._surface_points.df[
        geo_data._surface_points.df.series.isin(f_id)]
    # find neighbours
    neighbours = gp.select_nearest_surfaces_points(geo_data, fault_poi, 5)
    # calculate one fault orientation
    gp.set_orientation_from_neighbours(geo_data, neighbours[1])
    # find the calculated orientation
    test = geo_data._orientations.df.sort_index().iloc[-1][['dip', 'azimuth']].values

    # calculate reference
    reference = [90-np.arctan(0.5)/np.pi*180, 90]

    assert np.array_equal(reference, test)


def test_set_orientation_from_neighbours_all():
    data_path = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/'
    path_to_data = data_path + "/data/input_data/jan_models/"

    geo_data = gp.create_data('fault', extent=[0, 1000, 0, 1000, 0, 1000],
                              resolution=[50, 50, 50],
                          path_o=path_to_data + "model5_orientations.csv",
                          path_i=path_to_data + "model5_surface_points.csv")

    # count orientations before orientation calculation
    length_pre = geo_data._orientations.df.shape[0]

    # find neighbours
    neighbours = gp.select_nearest_surfaces_points(geo_data, geo_data._surface_points.df, 2)
    # calculate all fault orientations
    gp.set_orientation_from_neighbours_all(geo_data, neighbours)

    # count orientations after orientation calculation
    length_after = geo_data._orientations.df.shape[0]

    assert np.array_equal(geo_data._surface_points.df.shape[0],
                          length_after-length_pre)



import numpy as np
import pytest

import gempy_legacy as gp
import matplotlib as mpl
mpl.use('Agg')
import matplotlib.pyplot as plt
import os

# Input files
root = 'https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/data/input_data/turner_syncline/'
path = os.path.dirname(__file__) + '/../input_data/'
orientations_file = root + 'orientations_clean.csv'
contacts_file = root + 'contacts_clean.csv'
fp = path + 'dtm_rp.tif'
series_file = root + 'all_sorts_clean.csv'

bbox = (500000, 7490000, 545000, 7520000)
model_base = -1500  # Original 3200
model_top = 800

gdal = pytest.importorskip("gdal")


@pytest.fixture(scope='module')
def model():
    geo_model = gp.create_model('test_map2Loop')
    gp.init_data(
        geo_model,
        extent=[bbox[0], bbox[2], bbox[1], bbox[3], model_base, model_top],
        resolution=[50, 50, 80],
        path_o=orientations_file,
        path_i=contacts_file
    )

    # Load Topology

    geo_model.set_topography(source='gdal', filepath=fp)

    # Stack Processing
    contents = np.genfromtxt(series_file,
                             delimiter=',', dtype='U100')[1:, 4:-1]

    map_series_to_surfaces = {}
    for pair in contents:
        map_series_to_surfaces.setdefault(pair[1], []).append(pair[0])

    gp.map_stack_to_surfaces(geo_model, map_series_to_surfaces,
                             remove_unused_series=False)

    gp.plot_3d(geo_model, ve=None, show_topography=False
               , image=True, show_lith=False,
               kwargs_plot_data={'arrow_size': 300})

    return geo_model


def test_axial_anisotropy_type_data(model):
    geo_model = model

    geo_model._rescaling.toggle_axial_anisotropy()
    # gp.compute_model(geo_model, compute_mesh_options={'mask_topography': False})

    geo_model.surface_points.df[['X', 'Y', 'Z']] = geo_model.surface_points.df[['X_c', 'Y_c',
                                                                                'Z_c']]
    geo_model.orientations.df[['X', 'Y', 'Z']] = geo_model.orientations.df[['X_c', 'Y_c',
                                                                            'Z_c']]

    # This is a hack
    geo_model._grid.topography.extent = geo_model._grid.extent_c

    geo_model.set_regular_grid(geo_model._grid.extent_c, [50, 50, 50])

    gp.plot_3d(geo_model, ve=None, show_topography=False
               , image=True, show_lith=False,
               kwargs_plot_data={'arrow_size': 10})


def test_axial_anisotropy_type_extent(model):
    geo_model = model

    geo_model._rescaling.toggle_axial_anisotropy(type='extent')
    # gp.compute_model(geo_model, compute_mesh_options={'mask_topography': False})

    geo_model.surface_points.df[['X', 'Y', 'Z']] = geo_model.surface_points.df[['X_c', 'Y_c',
                                                                                'Z_c']]
    geo_model.orientations.df[['X', 'Y', 'Z']] = geo_model.orientations.df[['X_c', 'Y_c',
                                                                            'Z_c']]

    # This is a hack
    geo_model._grid.topography.extent = geo_model._grid.extent_c

    geo_model.set_regular_grid(geo_model._grid.extent_c, [50, 50, 50])

    gp.plot_3d(geo_model, ve=None, show_topography=False
               , image=True, show_lith=False,
               kwargs_plot_data={'arrow_size': 10})


def test_axial_anisotropy(model):
    # Location box

    geo_model = model
    geo_model._rescaling.toggle_axial_anisotropy()
    # gp.compute_model(geo_model, compute_mesh_options={'mask_topography': False})

    geo_model.surface_points.df[['X', 'Y', 'Z']] = geo_model.surface_points.df[['X_c', 'Y_c',
                                                                                'Z_c']]
    geo_model.orientations.df[['X', 'Y', 'Z']] = geo_model.orientations.df[['X_c', 'Y_c',
                                                                            'Z_c']]

    # This is a hack
    geo_model._grid.topography.extent = geo_model._grid.extent_c

    geo_model.set_regular_grid(geo_model._grid.extent_c, [50, 50, 50])
    geo_model.modify_kriging_parameters('range', 0.1)
    geo_model.modify_kriging_parameters('drift equations', [9, 9, 9, 9, 9])

    geo_model.modify_surface_points(
        geo_model.surface_points.df.index,
        smooth=0.001
    )

    gp.set_interpolator(geo_model, aesara_optimizer='fast_run', dtype='float64')
    gp.compute_model(geo_model, compute_mesh_options={'mask_topography': False,
                                                      'masked_marching_cubes': False})

    gp.plot_2d(geo_model,
               section_names=['topography'],
               show_topography=True,
               )
    plt.show()

    gp.plot_3d(geo_model, ve=None, show_topography=False, image=True, show_lith=False,
               kwargs_plot_data={'arrow_size': 10}
               )



# Importing GemPy
import gempy_legacy as gp


# Importing auxiliary libraries
import numpy as np

import gempy_legacy.core.grid


class TestGrid:
    def test_set_regular_grid(self):
        # Test creating an empty list
        grid = gempy_legacy.core.grid.Grid()

        # Test set regular grid by hand
        grid.create_regular_grid([0, 2000, 0, 2000, -2000, 0], [50, 50, 50])

    def test_grid_init(self):
        # Or we can init one of the default grids since the beginning by passing
        # the correspondant attributes
        grid = gempy_legacy.core.grid.Grid(extent=[0, 2000, 0, 2000, -2000, 0],
                                           resolution=[50, 50, 50])

    def test_section_grid(self):
        geo_data = gp.create_data('section_grid', [0, 1000, 0, 1000, 0, 1000], resolution=[10, 10, 10])
        geo_data.set_topography()
        section_dict = {'section1': ([0, 0], [1000, 1000], [100, 80]),
                        'section2': ([800, 0], [800, 1000], [150, 100]),
                        'section3': ([50, 200], [100, 500], [200, 150])}

        geo_data.set_section_grid(section_dict)

        print(geo_data._grid.sections)
        np.testing.assert_almost_equal(geo_data._grid.sections.df.loc['section3', 'dist'], 304.138127,
                                       decimal=4)

    def test_set_section_twice(self):
        geo_data = gp.create_data(extent=[0, 1000, 0, 1000, 0, 1000], resolution=[10, 10, 10])
        section_dict = {'section1': ([0, 0], [1000, 1000], [100, 80]),
                        'section2': ([800, 0], [800, 1000], [150, 100]),
                        'section3': ([50, 200], [100, 500], [200, 150])}

        geo_data.set_section_grid(section_dict)
        geo_data.set_section_grid(section_dict)
        print(geo_data._grid.sections)

    def test_custom_grid(self):
        # create custom grid
        grid = gempy_legacy.core.grid.Grid()
        cg = np.array([[1, 2, 3],
                        [4, 5, 6],
                        [7, 8, 9]])
        grid.create_custom_grid(cg)
        # make sure the custom grid is active
        assert grid.active_grids[1]
        # make sure the custom grid is equal to the provided values
        np.testing.assert_array_almost_equal(cg, grid.custom_grid.values)
        # make sure we have the correct number of values in our grid
        l0, l1 = grid.get_grid_args('custom')
        assert l0 == 0
        assert l1 == 3





import gempy_legacy.gempy_api as gp
import pytest


class TestModel:

    @pytest.fixture(scope='class')
    def test_create_model(self):
        model = gp.Project()
        print(model)
        return model




"""
Getting Started
===============

"""

# %%

# Importing GemPy
import gempy_legacy as gp

# Importing aux libraries
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.image as mpimg

# %%
# Initializing the model:
# ~~~~~~~~~~~~~~~~~~~~~~~
# 
# The first step to create a GemPy model is create a gempy.Model object
# that will contain all the other data structures and necessary
# functionality.
# 
# In addition for this example we will define a regular grid since the
# beginning. This is the grid where we will interpolate the 3D geological
# model. GemPy comes with an array of different grids for different
# pourposes as we will see below. For visualization usually a regular grid
# is the one that makes more sense.
# 

# %% 
geo_model = gp.create_model('Model1')
geo_model = gp.init_data(geo_model, extent=[0, 791, 0, 200, -582, 0], resolution=[100, 10, 100])

# %%
# GemPy core code is written in Python. However for efficiency (and other
# reasons) most of heavy computations happend in optimize compile code,
# either C or CUDA for GPU. To do so, GemPy rely on the library aesara. To
# guarantee maximum optimization aesara requires to compile the code for
# every Python kernel. The compilation is done by calling the following
# line at any point (before computing the model):
# 

# %% 
gp.set_interpolator(geo_model, aesara_optimizer='fast_compile', verbose=[])

# %%
# Creating figure:
# ~~~~~~~~~~~~~~~~
# 
# GemPy uses matplotlib and pyvista-vtk libraries for 2d and 3d
# visualization of the model respectively. One of the design decisions of
# GemPy is to allow real time construction of the model. What this means
# is that you can start adding input data and see in real time how the 3D
# surfaces evolve. Lets initialize the visualization windows.
# 
# The first one is the 2d figure. Just place the window where you can see
# it (maybe move the jupyter notebook to half screen and use the other
# half for the renderers).
# 

# %% 
# %matplotlib qt5
p2d = gp.plot_2d(geo_model)

# %%
# Add model section
# ^^^^^^^^^^^^^^^^^
# 
# In the 2d renderer we can add several cross section of the model. In
# this case, for simplicity sake we are just adding one perpendicular to
# y.
# 


# %%
# Loading cross-section image:
# ^^^^^^^^^^^^^^^^^^^^^^^^^^^^
# 
# Remember that gempy is simply using matplotlib and therofe the ax object
# created above is a standard matplotlib axes. This allow to manipulate it
# freely. Lets load an image with the information of couple of boreholes
# 

# %% 
# Reading image
img = mpimg.imread('wells.png')
# Plotting it inplace
p2d = gp.plot_2d(geo_model, show=False)
p2d.axes[0].imshow(img, origin='upper', alpha=.8, extent=(0, 791, -582, 0))
plt.show()

# %%
# We can do the same in 3D through pyvista and vtk rendering:
# 

# %% 
p3d = gp.plot_3d(geo_model)

# %%
# Building the model
# ------------------
# 
# Now that we have everything initialize we can start the construction of
# the geological model.
# 
# Surfaces
# ~~~~~~~~
# 
# GemPy is a surface based interpolator. This means that all the input
# data we add has to be refered to a surface. The surfaces always mark the
# bottom of a unit. By default GemPy surfaces are empty:
# 

# %% 
geo_model.surfaces

# %%
# If we do not care about the names and we just want to interpolate a
# surface we can use:
# 

# %% 
# Default surfaces:
geo_model.set_default_surfaces()

# %%
# Now we can start adding data. GemPy input data consist on surface points
# and orientations (perpendicular to the layers). The 2D plot gives you
# the X and Z coordinates when hovering the mouse over. We can add a
# surface point as follows:
# 

# %% 
# Add a point
geo_model.add_surface_points(X=223, Y=0.01, Z=-94, surface='surface1')

# Plot in 2D
gp.plot_2d(geo_model, cell_number=11)

# Plot in 3D
gp.plot_3d(geo_model)

# %%
# Now we can add the other two points of the layer:
# 

# %% 
# Add points
geo_model.add_surface_points(X=458, Y=0, Z=-107, surface='surface1')
geo_model.add_surface_points(X=612, Y=0, Z=-14, surface='surface1')

# Plotting
gp.plot_2d(geo_model, cell_number=11)
gp.plot_3d(geo_model)

# %%
# The minimum amount of data to interpolate anything in gempy is: a) 2
# surface points per surface b) One orientation per series.
# 
# Lets add an orientation anywhere in space:
# 

# %% 
# Adding orientation
geo_model.add_orientations(X=350, Y=0, Z=-300, surface='surface1', pole_vector=(0, 0, 1))
gp.plot_2d(geo_model, cell_number=5)
gp.plot_3d(geo_model)

# %%
# Now we have enough data for finally interpolate!
# 

# %% 
gp.compute_model(geo_model)

# %% 
geo_model.additional_data.kriging_data

# %% 
geo_model.additional_data.rescaling_data

# %%
# That is, we have interpolated the 3D surface. We can visualize:
# 

# %% 
# In 2D
gp.plot_2d(geo_model, cell_number=[5])

# In 3D
gp.plot_3d(geo_model, show_surfaces=True)

# %%
# Adding more layers:
# ~~~~~~~~~~~~~~~~~~~
# 
# So far we only need 2 units defined. The cross-section image that we
# load have 4 however. Lets add two layers more:
# 

# %% 
geo_model.surfaces

# %% 
geo_model.series

# %% 
geo_model.add_surfaces(['surface3', 'basement'])

# %%
# Layer 2
# ~~~~~~~
# 
# Add the layer next layers:
# 

# %% 
# Your code here:
geo_model.add_surface_points(X=225, Y=0, Z=-269, surface='surface2')
geo_model.add_surface_points(X=459, Y=0, Z=-279, surface='surface2')

# --------------------

# %% 
# Compute model
gp.compute_model(geo_model)

# %% 
gp.plot_2d(geo_model, cell_number=5, legend='force')
gp.plot_3d(geo_model)

# %%
# Layer 3
# ~~~~~~~
# 

# %% 
# Your code here:
geo_model.add_surface_points(X=225, Y=0, Z=-439, surface='surface3')
geo_model.add_surface_points(X=464, Y=0, Z=-456, surface='surface3')
geo_model.add_surface_points(X=619, Y=0, Z=-433, surface='surface3')

# ------------------

# %% 
# Computing and plotting 3D
gp.compute_model(geo_model)

gp.plot_2d(geo_model, cell_number=5, legend='force')
gp.plot_3d(geo_model, kwargs_plot_structured_grid={'opacity': .2})

# %%
# Adding a Fault
# --------------
# 
# So far the model is simply a depositional unit. GemPy allows for
# unconformities and faults to build complex models. This input is given
# by categorical data. In general:
# 
# input data (surface points/ orientations) <belong to< surface <belong
# to< series
# 
# And series can be a fault—i.e. offset the rest of surface— or not. We
# are going to show how to add a fault as an example.
# 
# First we need to add a series:
# 

# %% 
geo_model.add_features('Fault1')

# %% 
geo_model.reorder_features(['Fault1', 'Default series'])

# %%
# Then define that is a fault:
# 

# %% 
geo_model.set_is_fault('Fault1')

# %%
# But we also need to add a new surface:
# 

# %% 
geo_model.add_surfaces('fault1')

# %%
# And finally assign the new surface to the new series/fault
# 

# %% 
gp.map_stack_to_surfaces(geo_model, {'Fault1': 'fault1'})

# %%
# Now we can just add input data as before (remember the minimum amount of
# input data to compute a model):
# 

# %% 
# Add input data of the fault
geo_model.add_surface_points(X=550, Y=0, Z=-30, surface='fault1')
geo_model.add_surface_points(X=650, Y=0, Z=-200, surface='fault1')
geo_model.add_orientations(X=600, Y=0, Z=-100, surface='fault1', pole_vector=(.3, 0, .3))

# Plotting Inpute data
gp.plot_2d(geo_model, show_solutions=False)

# %%
# And now is computing as before:
# 

# %% 
# Compute
gp.compute_model(geo_model)

# Plot
gp.plot_2d(geo_model, cell_number=5, legend='force')
gp.plot_3d(geo_model, kwargs_plot_structured_grid={'opacity': .2})

# %%
# As you can see now instead of having folding layers we have a sharp
# jump. Building on this you can pretty much any model you can imagine.
# 


# %%
# Additional features:
# ====================
# 
# Over the years we have built a bunch of assets integrate with gempy.
# Here we will show some of them:
# 
# Topography
# ----------
# 
# GemPy has a built-in capabilities to read and manipulate topographic
# data (through gdal). To show an example we can just create a random
# topography:
# 

# %% 
# Adding random topography
geo_model.set_topography(source='random', fd=1.9, d_z=np.array([-150, 0]),
                         resolution=np.array([200, 200]))

# %%
# The topography can we visualize in both renderers:
# 

# %% 
gp.plot_2d(geo_model, cell_number=5, legend='force')
gp.plot_3d(geo_model, kwargs_plot_structured_grid={'opacity':.2})

# %%
# But also allows us to compute the geological map of an area:
# 

# %% 
gp.compute_model(geo_model)

# sphinx_gallery_thumbnail_number = 16
gp.plot_3d(geo_model, show_topography=True)

# %%
# Gravity inversion
# -----------------
# 
# GemPy also allows for inversions (in production only gravity so far). We
# can see a small demo how this works.
# 
# The first thing to do is to assign densities to each of the units:
# 

# %% 
geo_model.add_surface_values([0, 2.6, 2.4, 3.2, 3.6], ['density'])

# %%
# Also we can create a centered grid around a device for precision:
# 

# %% 
geo_model.set_centered_grid(centers=[[400, 0, 0]], resolution=[10, 10, 100], radius=800)

# %%
# We need to modify the compile code:
# 

# %% 
gp.set_interpolator(geo_model, output=['gravity'], aesara_optimizer='fast_run')

# %%
# But now additionally to the interpolation we also compute the forward
# gravity of the model (at the point XYZ = 400, 0, 0)
# 

# %% 
gp.compute_model(geo_model)
geo_model.solutions.fw_gravity



"""
2.1 Forward Gravity: Simple example
^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^

"""

# %%
# Importing gempy
import gempy_legacy as gp
from gempy_legacy.assets.geophysics import GravityPreprocessing

# Aux imports
import numpy as np
import pandas as pd
import os
import matplotlib.pyplot as plt

np.random.seed(1515)
pd.set_option('display.precision', 2)

# %%
cwd = os.getcwd()
if 'examples' not in cwd:
    data_path = os.getcwd() + '/examples/'
else:
    data_path = cwd + '/../../'

geo_model = gp.load_model('Greenstone', path=data_path + 'data/gempy_models/Greenstone')

# %% 
geo_model.stack

# %% 
geo_model.surfaces

# %% 
gp.plot_2d(geo_model)

# %%

# %%
# Creating grid
# ~~~~~~~~~~~~~
# 

# %%
# First we need to define the location of the devices. For this example we
# can make a map:
# 

# %% 
grav_res = 20
X = np.linspace(7.050000e+05, 747000, grav_res)
Y = np.linspace(6863000, 6925000, grav_res)
Z = 300
xyz = np.meshgrid(X, Y, Z)
xy_ravel = np.vstack(list(map(np.ravel, xyz))).T
xy_ravel

# %%
# We can see the location of the devices relative to the model data:
# 

# %%

gp.plot_2d(geo_model, direction='z', show=False)
plt.scatter(xy_ravel[:, 0], xy_ravel[:, 1], s=1)
plt.show()

# %%
# Now we need to create the grid centered on the devices (see:
# https://github.com/cgre-aachen/gempy/blob/master/notebooks/tutorials/ch1-3-Grids.ipynb)
# 

# %% 
geo_model.set_centered_grid(xy_ravel, resolution=[10, 10, 15], radius=5000)

# %% 
geo_model.grid.centered_grid.kernel_centers

# %%
# Now we need to compute the component tz (see
# https://github.com/cgre-achen/gempy/blob/master/notebooks/tutorials/ch2-2-Cell_selection.ipynb)
# 

# %% 
g = GravityPreprocessing(geo_model.grid.centered_grid)

# %% 
tz = g.set_tz_kernel()

# %% 
tz

# %%
# Compiling the gravity graph
# ~~~~~~~~~~~~~~~~~~~~~~~~~~~
# 
# If geo_model has already a centered grid, the calculation of tz happens
# automatically.  This aesara graph will return gravity
# as well as the lithologies. In addition we need either to pass the density
# block (see below). Or the position of density on the surface(in the
# future the name) to compute the density block at running time.
# 

# %% 
geo_model.surfaces

# %%
# In this case the densities of each layer are at the loc 1 (0 is the id)
# 

# New way
gp.set_interpolator(geo_model, output=['gravity'], pos_density=1, gradient=False,
                    aesara_optimizer='fast_run')

# %%
# Once we have created a gravity interpolator we can call it from compute
# model as follows:
# 

# %% 
sol = gp.compute_model(geo_model)
grav = sol.fw_gravity

# %% 
gp.plot_2d(geo_model, direction=['z'], height=7, show_results=False, show_data=True,
           show=False)
plt.scatter(xy_ravel[:, 0], xy_ravel[:, 1], s=1)
plt.imshow(sol.fw_gravity.reshape(grav_res, grav_res),
           extent=(xy_ravel[:, 0].min() + (xy_ravel[0, 0] - xy_ravel[1, 0]) / 2,
                   xy_ravel[:, 0].max() - (xy_ravel[0, 0] - xy_ravel[1, 0]) / 2,
                   xy_ravel[:, 1].min() + (xy_ravel[0, 1] - xy_ravel[30, 1]) / 2,
                   xy_ravel[:, 1].max() - (xy_ravel[0, 1] - xy_ravel[30, 1]) / 2),
           cmap='viridis_r', origin='lower')
plt.show()
# %%
# Plotting lithologies
# ^^^^^^^^^^^^^^^^^^^^
# 
# If we want to compute the lithologies we will need to create a normal
# interpolator object as seen in the Chapter 1 of the tutorials
# 


# %%
# Now we can plot all together (change the alpha parameter to see the
# gravity overlying):
# 

# %%
# sphinx_gallery_thumbnail_number = 4
gp.plot_2d(geo_model, cell_number=[-1], direction=['z'], show=False,
           kwargs_regular_grid={'alpha': .5})

plt.scatter(xy_ravel[:, 0], xy_ravel[:, 1], s=1)
plt.imshow(grav.reshape(grav_res, grav_res),
           extent=(xy_ravel[:, 0].min() + (xy_ravel[0, 0] - xy_ravel[1, 0]) / 2,
                   xy_ravel[:, 0].max() - (xy_ravel[0, 0] - xy_ravel[1, 0]) / 2,
                   xy_ravel[:, 1].min() + (xy_ravel[0, 1] - xy_ravel[30, 1]) / 2,
                   xy_ravel[:, 1].max() - (xy_ravel[0, 1] - xy_ravel[30, 1]) / 2),
           cmap='viridis_r', origin='lower', alpha=.8)
cbar = plt.colorbar()
cbar.set_label(r'$\mu$gal')
plt.show()



import pytest

import aesara
import numpy as np
import sys, os

import gempy_legacy.core.data_modules.stack

sys.path.append("../..")
import gempy_legacy as gp


input_path = os.path.dirname(__file__) + '/../../examples/data'









"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


Module with classes and methods to visualized structural geology data and potential fields of the regional modelling based on
the potential field method.

Created on 23/09/2019

@author: Miguel de la Varga, Elisa Heim
"""

import warnings
import os
import copy
import numpy as np
import matplotlib.pyplot as plt
import matplotlib.colors as mcolors
from matplotlib.ticker import FixedFormatter, FixedLocator
import matplotlib.gridspec as gridspect
import matplotlib as mpl
import scipy.spatial.distance as dd
import seaborn as sns

sns.set_context('talk')
plt.style.use(['seaborn-v0_8-white', 'seaborn-v0_8-talk'])

warnings.filterwarnings("ignore", message="No contour levels were found")


class Plot2D:
    """
    Class with functionality to plot 2D gempy sections

    Args:
        model: gempy.Model object
        cmap: Color map to pass to matplotlib
    """

    def __init__(self, model, cmap=None, norm=None, **kwargs):
        self.model = model
        self._color_lot = dict(zip(self.model._surfaces.df['surface'],
                                   self.model._surfaces.df['color']))
        self.axes = list()

        if cmap is None:
            self.cmap = mcolors.ListedColormap(list(self.model._surfaces.df['color']))
            self._custom_colormap = False
        else:
            self.cmap = cmap
            self._custom_colormap = True

        if norm is None:
            self.norm = mcolors.Normalize(vmin=0.5, vmax=len(self.cmap.colors) + 0.5)
        else:
            self.norm = norm

    def update_colot_lot(self, color_dir=None):
        if color_dir is None:
            color_dir = dict(zip(self.model._surfaces.df['surface'], self.model._surfaces.df['color']))

        self._color_lot = color_dir
        if self._custom_colormap is False:
            self.cmap = mcolors.ListedColormap(list(self.model._surfaces.df['color']))
            self.norm = mcolors.Normalize(vmin=0.5, vmax=len(self.cmap.colors) + 0.5)

    @staticmethod
    def remove(ax):
        while len(ax.collections) != 0:
            list(map(lambda x: x.remove(), ax.collections))

    def _make_section_xylabels(self, section_name, n=5):
        """
        @elisa heim
        Setting the axis labels to any combination of vertical crossections

        Args:
            section_name: name of a defined gempy crossection. See gempy.Model().grid.section
            n:

        Returns:

        """
        if n > 5:
            n = 3  # todo I don't know why but sometimes it wants to make a lot of xticks
        elif n < 0:
            n = 3

        j = np.where(self.model._grid.sections.names == section_name)[0][0]
        startend = list(self.model._grid.sections.section_dict.values())[j]
        p1, p2 = startend[0], startend[1]
        xy = self.model._grid.sections.calculate_line_coordinates_2points(p1, p2, n)
        if len(np.unique(xy[:, 0])) == 1:
            labels = xy[:, 1].astype(int)
            axname = 'Y'
        elif len(np.unique(xy[:, 1])) == 1:
            labels = xy[:, 0].astype(int)
            axname = 'X'
        else:
            labels = [str(xy[:, 0].astype(int)[i]) + ',\n' + str(xy[:, 1].astype(int)[i]) for i in
                      range(xy[:, 0].shape[0])]
            axname = 'X,Y'
        return labels, axname

    def _slice(self, direction, cell_number=25):
        """
        Slice the 3D array (blocks or scalar field) in the specific direction selected in the plot functions

        """
        _a, _b, _c = (slice(0, self.model._grid.regular_grid.resolution[0]),
                      slice(0, self.model._grid.regular_grid.resolution[1]),
                      slice(0, self.model._grid.regular_grid.resolution[2]))

        if direction == "x":
            cell_number = int(self.model._grid.regular_grid.resolution[0] / 2) if cell_number == 'mid' else cell_number
            _a, x, y, Gx, Gy = cell_number, "Y", "Z", "G_y", "G_z"
            extent_val = self.model._grid.regular_grid.extent[[2, 3, 4, 5]]
        elif direction == "y":
            cell_number = int(self.model._grid.regular_grid.resolution[1] / 2) if cell_number == 'mid' else cell_number
            _b, x, y, Gx, Gy = cell_number, "X", "Z", "G_x", "G_z"
            extent_val = self.model._grid.regular_grid.extent[[0, 1, 4, 5]]
        elif direction == "z":
            cell_number = int(self.model._grid.regular_grid.resolution[2] / 2) if cell_number == 'mid' else cell_number
            _c, x, y, Gx, Gy = cell_number, "X", "Y", "G_x", "G_y"
            extent_val = self.model._grid.regular_grid.extent[[0, 1, 2, 3]]
        else:
            raise AttributeError(str(direction) + "must be a cartesian direction, i.e. xyz")
        return _a, _b, _c, extent_val, x, y, Gx, Gy

    def create_figure(self, figsize=None, textsize=None, **kwargs):
        """
        Create the figure.

        Args:
            figsize:
            textsize:

        Returns:
            figure, list axes, subgrid values
        """
        cols = kwargs.get('cols', 1)
        rows = kwargs.get('rows', 1)

        figsize, self.ax_labelsize, _, self.xt_labelsize, self.linewidth, _ = _scale_fig_size(
            figsize, textsize, rows, cols)
        self.fig = plt.figure(figsize=figsize, constrained_layout=False)
        self.fig.is_legend = False
        # TODO make grid variable
        # self.gs_0 = gridspect.GridSpec(2, 2, figure=self.fig, hspace=.9)

        return self.fig, self.axes  # , self.gs_0

    def add_section(self, section_name=None, cell_number=None, direction='y', ax=None, ax_pos=111,
                    ve=1., **kwargs):

        extent_val = kwargs.get('extent', None)
        self.update_colot_lot()

        if ax is None:
            ax = self.fig.add_subplot(ax_pos)

        if section_name is not None:
            if section_name == 'topography':
                ax.set_title('Geological map')
                ax.set_xlabel('X')
                ax.set_ylabel('Y')
                extent_val = self.model._grid.topography.extent
            else:

                dist = self.model._grid.sections.df.loc[section_name, 'dist']

                extent_val = [0, dist,
                              self.model._grid.regular_grid.extent[4], self.model._grid.regular_grid.extent[5]]

                labels, axname = self._make_section_xylabels(section_name, len(ax.get_xticklabels()) - 2)
                pos_list = np.linspace(0, dist, len(labels))
                ax.xaxis.set_major_locator(FixedLocator(nbins=len(labels), locs=pos_list))
                ax.xaxis.set_major_formatter(FixedFormatter((labels)))
                ax.set(title=section_name, xlabel=axname, ylabel='Z')

        elif cell_number is not None:
            _a, _b, _c, extent_val, x, y = self._slice(direction, cell_number)[:-2]
            ax.set_xlabel(x)
            ax.set_ylabel(y)
            ax.set(title='Cell Number: ' + str(cell_number) + ' Direction: ' + str(direction))

        if extent_val is not None:
            if extent_val[3] < extent_val[2]:  # correct vertical orientation of plot
                ax.invert_yaxis()
            self._aspect = (extent_val[3] - extent_val[2]) / (extent_val[1] - extent_val[0]) / ve
            ax.set_xlim(extent_val[0], extent_val[1])
            ax.set_ylim(extent_val[2], extent_val[3])
        ax.set_aspect('equal')

        # Adding some properties to the axes to make easier to plot
        ax.section_name = section_name
        ax.cell_number = cell_number
        ax.direction = direction
        ax.tick_params(axis='x', labelrotation=30)
        self.axes = np.append(self.axes, ax)
        self.fig.tight_layout()

        return ax

    @staticmethod
    def _check_default_section(ax, section_name, cell_number, direction):

        if section_name is None:
            try:
                section_name = ax.section_name
            except AttributeError:
                pass
        if cell_number is None:
            try:
                cell_number = ax.cell_number
                direction = ax.direction
            except AttributeError:
                pass

        return section_name, cell_number, direction

    def plot_regular_grid(self, ax, section_name=None, cell_number=None, direction='y',
                          block: np.ndarray = None, resolution=None, **kwargs):
        """Generic function to plot all regular data

        Args:
            block:
            section_name:
            cell_number:
            direction:
            ax:
            **kwargs: imshow kwargs

        Returns:

        """
        self.update_colot_lot()
        extent_val = [*ax.get_xlim(), *ax.get_ylim()]
        if 'cmap' in kwargs:
            cmap = kwargs['cmap']
        else:
            cmap = self.cmap
        if 'norm' in kwargs:
            norm = kwargs['norm']
        else:
            norm = self.norm

        section_name, cell_number, direction = self._check_default_section(ax, section_name, cell_number, direction)

        if section_name is not None:
            if section_name == 'topography':
                try:
                    image = self.model.solutions.geological_map[0].reshape(
                        self.model._grid.topography.values_2d[:, :, 2].shape).T

                except AttributeError:
                    raise AttributeError('Geological map not computed. Activate the topography grid.')
            else:
                assert type(section_name) == str or type(
                    section_name) == np.str_, 'section name must be a string of the name of the section'
                assert self.model.solutions.sections is not None, 'no sections for plotting defined'

                l0, l1 = self.model._grid.sections.get_section_args(section_name)
                shape = self.model._grid.sections.df.loc[section_name, 'resolution']
                image = self.model.solutions.sections[0][l0:l1].reshape(shape[0], shape[1]).T

        elif cell_number is not None or block is not None:
            _a, _b, _c, _, x, y = self._slice(direction, cell_number)[:-2]
            
            plot_block = block.reshape(self.model._grid.regular_grid.resolution)
            image = plot_block[_a, _b, _c].T
        else:
            raise AttributeError

        ax.imshow(image, origin='lower', zorder=-100,
                  cmap=cmap, norm=norm, extent=extent_val)
        return ax

    def plot_lith(self, ax, section_name=None, cell_number=None, direction='y', **kwargs):
        block = self.model.solutions.lith_block
        self.plot_regular_grid(ax, section_name, cell_number, direction, block=block)

    def plot_values(self, ax, series_n=0, section_name=None, cell_number=None,
                    direction='y', **kwargs):
        block = self.model.solutions.values_matrix[series_n]
        self.plot_regular_grid(ax, section_name, cell_number, direction, block=block,
                               **kwargs)

    def plot_block(self, ax, series_n=0, section_name=None, cell_number=None, direction='y',
                   **kwargs):

        block = self.model.solutions.block_matrix[series_n]
        self.plot_regular_grid(ax, section_name, cell_number, direction, block=block)

    def plot_scalar_field(self, ax, section_name=None, cell_number=None, series_n=0, direction='y',
                          block=None, **kwargs):
        """
        Plot the scalar field of a section.

        Args:
            ax:
            section_name:
            cell_number:
            series_n:
            direction:
            block:
            **kwargs:

        Returns:

        """

        extent_val = [*ax.get_xlim(), *ax.get_ylim()]
        section_name, cell_number, direction = self._check_default_section(ax, section_name, cell_number, direction)

        if section_name is not None:
            if section_name == 'topography':
                try:
                    image = self.model.solutions.geological_map[1][series_n].reshape(
                        self.model._grid.topography.values_3D[:, :, 2].shape)
                except AttributeError:
                    raise AttributeError('Geological map not computed. Activate the topography grid.')
            else:
                l0, l1 = self.model._grid.sections.get_section_args(section_name)
                shape = self.model._grid.sections.df.loc[section_name, 'resolution']
                image = self.model.solutions.sections[1][series_n][l0:l1].reshape(shape).T

        elif cell_number is not None or block is not None:
            _a, _b, _c, _, x, y = self._slice(direction, cell_number)[:-2]
            if block is None:
                _block = self.model.solutions.scalar_field_matrix[series_n]
            else:
                _block = block

            plot_block = _block.reshape(self.model._grid.regular_grid.resolution)
            image = plot_block[_a, _b, _c].T
        else:
            raise AttributeError

        ax.contour(image, cmap='autumn', extent=extent_val, zorder=8, **kwargs)
        if 'N' in kwargs:
            kwargs.pop('N')
        ax.contourf(image, cmap='autumn', extent=extent_val, zorder=7, alpha=.8,
                    **kwargs)

    def plot_data(self, ax, section_name=None, cell_number=None, direction='y',
                  legend=True,
                  projection_distance=None, **kwargs):
        """
        Plot data--i.e. surface_points and orientations--of a section.

        Args:
            ax:
            section_name:
            cell_number:
            direction:
            legend: bool or 'force'
            projection_distance:
            **kwargs:

        Returns:

        """
        if projection_distance is None:
            projection_distance = 0.2 * self.model._rescaling.df['rescaling factor'].values[0]

        self.update_colot_lot()

        points = self.model._surface_points.df.copy()
        orientations = self.model._orientations.df.copy()
        section_name, cell_number, direction = self._check_default_section(ax, section_name, cell_number, direction)

        if section_name is not None:
            if section_name == 'topography':
                topo_comp = kwargs.get('topo_comp', 5000)
                decimation_aux = int(self.model._grid.topography.values.shape[0] / topo_comp)
                tpp = self.model._grid.topography.values[::decimation_aux + 1, :]
                
                cdist_sp = dd.cdist(tpp, self.model._surface_points.df[['X', 'Y', 'Z']])
                cartesian_point_dist = (cdist_sp < projection_distance).sum(axis=0).astype(bool)

                cdist_ori = dd.cdist(tpp, self.model._orientations.df[['X', 'Y', 'Z']])
                cartesian_ori_dist = (cdist_ori < projection_distance).sum(axis=0).astype(bool)

                x, y, Gx, Gy = 'X', 'Y', 'G_x', 'G_y'
            else:
                # Project points:
                shift = np.asarray(self.model._grid.sections.df.loc[section_name, 'start'])
                end_point = np.atleast_2d(np.asarray(self.model._grid.sections.df.loc[section_name, 'stop']) - shift)
                A_rotate = np.dot(end_point.T, end_point) / self.model._grid.sections.df.loc[section_name, 'dist'] ** 2

                perpe_sqdist = ((np.dot(A_rotate, (points[['X', 'Y']]).T).T - points[['X', 'Y']]) ** 2).sum(axis=1)
                cartesian_point_dist = np.sqrt(perpe_sqdist)

                cartesian_ori_dist = np.sqrt(((np.dot(
                    A_rotate, (orientations[['X', 'Y']]).T).T - orientations[['X', 'Y']]) ** 2).sum(axis=1))

                # This are the coordinates of the data projected on the section
                cartesian_point = np.dot(A_rotate, (points[['X', 'Y']] - shift).T).T
                cartesian_ori = np.dot(A_rotate, (orientations[['X', 'Y']] - shift).T).T

                # Since we plot only the section we want the norm of those coordinates
                points['X'] = np.linalg.norm(cartesian_point, axis=1)
                orientations['X'] = np.linalg.norm(cartesian_ori, axis=1)
                x, y, Gx, Gy = 'X', 'Z', 'G_x', 'G_z'

        else:

            if cell_number is None:
                cell_number = int(self.model._grid.regular_grid.resolution[0] / 2)
            elif cell_number == 'mid':
                cell_number = int(self.model._grid.regular_grid.resolution[0] / 2)

            if direction == 'x' or direction == 'X':
                arg_ = 0
                dx = self.model._grid.regular_grid.dx
                dir = 'X'
            elif direction == 'y' or direction == 'Y':
                arg_ = 2
                dx = self.model._grid.regular_grid.dy
                dir = 'Y'
            elif direction == 'z' or direction == 'Z':
                arg_ = 4
                dx = self.model._grid.regular_grid.dz
                dir = 'Z'
            else:
                raise AttributeError('Direction must be x, y, z')

            _loc = self.model._grid.regular_grid.extent[arg_] + dx * cell_number
            cartesian_point_dist = points[dir] - _loc
            cartesian_ori_dist = orientations[dir] - _loc

            x, y, Gx, Gy = self._slice(direction)[4:]

        select_projected_p = cartesian_point_dist < projection_distance
        select_projected_o = cartesian_ori_dist < projection_distance

        # Hack to keep the right X label:
        temp_label = copy.copy(ax.xaxis.label)

        points_df = points[select_projected_p]
        
        _colors = points_df['surface'].map(self._color_lot)
        points_df['colors'] = _colors

        points_df.plot.scatter(x=x, y=y, ax=ax,
                               c=_colors,
                               s=70, 
                               zorder=102,
                               edgecolors='white',
                               colorbar=False
                               )
        
        if self.fig.is_legend is False and legend is True or legend == 'force':
            markers = [plt.Line2D([0, 0], [0, 0], color=color, marker='o', linestyle='') for color in self._color_lot.values()]
            ax.legend(markers, self._color_lot.keys(), numpoints=1)
            self.fig.is_legend = True
        ax.xaxis.label = temp_label

        sel_ori = orientations[select_projected_o]

        aspect = np.subtract(*ax.get_ylim()) / np.subtract(*ax.get_xlim())
        min_axis = 'width' if aspect < 1 else 'height'

        # Eli options
        ax.quiver(sel_ori[x], sel_ori[y], sel_ori[Gx], sel_ori[Gy],
                  pivot="tail", scale_units=min_axis, scale=30, color=sel_ori['surface'].map(self._color_lot),
                  edgecolor='k', headwidth=8, linewidths=1, zorder=102)

        try:
            ax.legend_.set_frame_on(True)
            ax.legend_.set_zorder(10000)
        except AttributeError:
            pass

    def calculate_p1p2(self, direction, cell_number):

        if direction == 'y':
            cell_number = int(self.model._grid.regular_grid.resolution[1] / 2) if cell_number == 'mid' else cell_number

            y = self.model._grid.regular_grid.extent[2] + self.model._grid.regular_grid.dy * cell_number
            p1 = [self.model._grid.regular_grid.extent[0], y]
            p2 = [self.model._grid.regular_grid.extent[1], y]

        elif direction == 'x':
            cell_number = int(self.model._grid.regular_grid.resolution[0] / 2) if cell_number == 'mid' else cell_number

            x = self.model._grid.regular_grid.extent[0] + self.model._grid.regular_grid.dx * cell_number
            p1 = [x, self.model._grid.regular_grid.extent[2]]
            p2 = [x, self.model._grid.regular_grid.extent[3]]

        else:
            raise NotImplementedError
        return p1, p2

    def _slice_topo_4_sections(self, p1, p2, resx, method='interp2d'):
        """
        Slices topography along a set linear section

        Args:
            :param p1: starting point (x,y) of the section
            :param p2: end point (x,y) of the section
            :param resx: resolution of the defined section
            :param method: interpolation method, 'interp2d' for cubic scipy.interpolate.interp2d
                                             'spline' for scipy.interpolate.RectBivariateSpline

        Returns:
            :return: returns x,y,z values of the topography along the section
        """
        xy = self.model._grid.sections.calculate_line_coordinates_2points(p1, p2, resx)
        z = self.model._grid.sections.interpolate_zvals_at_xy(xy, self.model._grid.topography, method)
        return xy[:, 0], xy[:, 1], z

    def plot_topography(self, ax, fill_contour=False,
                        contour=True,
                        section_name=None,
                        cell_number=None, direction='y', block=None, **kwargs):

        hillshade = kwargs.get('hillshade', True)
        azdeg = kwargs.get('azdeg', 0)
        altdeg = kwargs.get('altdeg', 0)
        cmap = kwargs.get('cmap', 'terrain')

        self.update_colot_lot()
        section_name, cell_number, direction = self._check_default_section(ax, section_name, cell_number, direction)

        if section_name is not None and section_name != 'topography':

            p1 = self.model._grid.sections.df.loc[section_name, 'start']
            p2 = self.model._grid.sections.df.loc[section_name, 'stop']
            x, y, z = self._slice_topo_4_sections(p1, p2, self.model._grid.topography.resolution[0])

            pseudo_x = np.linspace(0, self.model._grid.sections.df.loc[section_name, 'dist'], z.shape[0])
            a = np.vstack((pseudo_x, z)).T
            xy = np.append(a,
                           ([self.model._grid.sections.df.loc[section_name, 'dist'], a[:, 1][-1]],
                            [self.model._grid.sections.df.loc[section_name, 'dist'],
                             self.model._grid.regular_grid.extent[5]],
                            [0, self.model._grid.regular_grid.extent[5]],
                            [0, a[:, 1][0]])).reshape(-1, 2)

            ax.fill(xy[:, 0], xy[:, 1], 'k', zorder=10)

        elif section_name == 'topography':
            
            import skimage
            from gempy_legacy.plot.helpers import add_colorbar
            topo = self.model._grid.topography
            topo_super_res = skimage.transform.resize(
                topo.values_2d,
                (1600, 1600),
                order=3,
                mode='edge',
                anti_aliasing=True, preserve_range=False)

            values = topo_super_res[:, :, 2].T
            if contour is True:
                CS = ax.contour(values, extent=(topo.extent[:4]),
                                colors='k', linestyles='solid', origin='lower')
                ax.clabel(CS, inline=1, fontsize=10, fmt='%d')
            if fill_contour is True:
                CS2 = ax.contourf(values, extent=(topo.extent[:4]), cmap=cmap)
                add_colorbar(axes=ax, label='elevation [m]', cs=CS2)

            if hillshade is True:
                from matplotlib.colors import LightSource

                ls = LightSource(azdeg=azdeg, altdeg=altdeg)
                hillshade_topography = ls.hillshade(values)
                ax.imshow(hillshade_topography, origin='lower', extent=topo.extent[:4], alpha=0.5, zorder=11,
                          cmap='gray')

        elif cell_number is not None or block is not None:
            p1, p2 = self.calculate_p1p2(direction, cell_number)
            resx = self.model._grid.regular_grid.resolution[0]
            resy = self.model._grid.regular_grid.resolution[1]

            try:
                x, y, z = self._slice_topo_4_sections(p1, p2, resx)
                if direction == 'x':
                    a = np.vstack((y, z)).T
                    ext = self.model._grid.regular_grid.extent[[2, 3]]
                elif direction == 'y':
                    a = np.vstack((x, z)).T
                    ext = self.model._grid.regular_grid.extent[[0, 1]]
                else:
                    raise NotImplementedError
                a = np.append(a,
                              ([ext[1], a[:, 1][-1]],
                               [ext[1], self.model._grid.regular_grid.extent[5]],
                               [ext[0], self.model._grid.regular_grid.extent[5]],
                               [ext[0], a[:, 1][0]]))
                line = a.reshape(-1, 2)
                ax.fill(line[:, 0], line[:, 1], color='k')
            except IndexError:
                warnings.warn('Topography needs to be a raster to be able to plot it'
                              'in 2D sections')
        return ax

    def plot_contacts(self, ax, section_name=None, cell_number=None, direction='y', block=None,
                      only_faults=False, **kwargs):
        self.update_colot_lot()
        section_name, cell_number, direction = self._check_default_section(ax, section_name, cell_number, direction)

        if only_faults:
            contour_idx = list(self.model._faults.df[self.model._faults.df['isFault'] == True].index)
        else:
            contour_idx = list(self.model._surfaces.df.index)
        extent_val = [*ax.get_xlim(), *ax.get_ylim()]
        zorder = kwargs.get('zorder', 100)

        if section_name is not None:
            if section_name == 'topography':
                shape = self.model._grid.topography.resolution

                scalar_fields = self.model.solutions.geological_map[1:]
                c_id = 0  # color id startpoint

                for e, block in enumerate(scalar_fields):
                    level = self.model.solutions.scalar_field_at_surface_points[e][np.where(
                        self.model.solutions.scalar_field_at_surface_points[e] != 0)]

                    c_id2 = c_id + len(level)  # color id endpoint
                    ax.contour(
                        block.reshape(shape).T, 0,
                        levels=np.sort(level),
                        colors=self.cmap.colors[c_id:c_id2][::-1],
                        linestyles='solid',
                        origin='lower',
                        extent=extent_val,
                        zorder=zorder - (e + len(level))
                    )
                    c_id = c_id2

            else:
                l0, l1 = self.model._grid.sections.get_section_args(section_name)
                shape = self.model._grid.sections.df.loc[section_name, 'resolution']
                scalar_fields = self.model.solutions.sections[1:][:, l0:l1]

                c_id = 0  # color id startpoint

                for e, block in enumerate(scalar_fields):
                    level = self.model.solutions.scalar_field_at_surface_points[e][np.where(
                        self.model.solutions.scalar_field_at_surface_points[e] != 0)]

                    # Ignore warning about some scalars not being on the plot since it is very common
                    # that an interface does not exit for a given section
                    c_id2 = c_id + len(level)  # color id endpoint
                    color_list = self.model._surfaces.df.groupby('isActive').get_group(True)['color'][c_id:c_id2][::-1]

                    ax.contour(block.reshape(shape).T, 0, levels=np.sort(level),
                               # colors=self.cmap.colors[self.model.surfaces.df['isActive']][c_id:c_id2],
                               colors=color_list,
                               linestyles='solid', origin='lower',
                               extent=extent_val, zorder=zorder - (e + len(level))
                               )
                    c_id = c_id2

        elif cell_number is not None or block is not None:
            _slice = self._slice(direction, cell_number)[:3]
            shape = self.model._grid.regular_grid.resolution
            c_id = 0  # * color id startpoint

            for e, block in enumerate(self.model.solutions.scalar_field_matrix):
                level = self.model.solutions.scalar_field_at_surface_points[e][np.where(
                    self.model.solutions.scalar_field_at_surface_points[e] != 0)]
                c_id2 = c_id + len(level)

                color_list = self.model._surfaces.df.groupby('isActive').get_group(True)['color'][c_id:c_id2][::-1]
                #    print(color_list)

                ax.contour(block.reshape(shape)[_slice].T, 0, levels=np.sort(level),
                           colors=color_list,
                           linestyles='solid', origin='lower',
                           extent=extent_val, zorder=zorder - (e + len(level))
                           )
                c_id = c_id2

    def plot_section_traces(self, ax, section_names=None, show_data=True, **kwargs):

        if section_names is None:
            section_names = list(self.model._grid.sections.names)

        if show_data:
            self.plot_data(ax, section_name='topography', **kwargs)

        for section in section_names:
            j = np.where(self.model._grid.sections.names == section)[0][0]
            x1, y1 = np.asarray(self.model._grid.sections.df.loc[section, 'start'])
            x2, y2 = np.asarray(self.model._grid.sections.df.loc[section, 'stop'])
            ax.plot([x1, x2], [y1, y2], label=section, linestyle='--')
            ax.legend(frameon=True)

    def plot_topo_g(self, ax, G, centroids, direction="y",
                    label_kwargs=None, node_kwargs=None, edge_kwargs=None):
        res = self.model._grid.regular_grid.resolution
        if direction == "y":
            c1, c2 = (0, 2)
            e1 = self.model._grid.regular_grid.extent[1] - self.model._grid.regular_grid.extent[0]
            e2 = self.model._grid.regular_grid.extent[5] - self.model._grid.regular_grid.extent[4]
            d1 = self.model._grid.regular_grid.extent[0]
            d2 = self.model._grid.regular_grid.extent[4]
            if len(list(centroids.items())[0][1]) == 2:
                c1, c2 = (0, 1)
            r1 = res[0]
            r2 = res[2]
        elif direction == "x":
            c1, c2 = (1, 2)
            e1 = self.model._grid.regular_grid.extent[3] - self.model._grid.regular_grid.extent[2]
            e2 = self.model._grid.regular_grid.extent[5] - self.model._grid.regular_grid.extent[4]
            d1 = self.model._grid.regular_grid.extent[2]
            d2 = self.model._grid.regular_grid.extent[4]
            if len(list(centroids.items())[0][1]) == 2:
                c1, c2 = (0, 1)
            r1 = res[1]
            r2 = res[2]
        elif direction == "z":
            c1, c2 = (0, 1)
            e1 = self.model._grid.regular_grid.extent[1] - self.model._grid.regular_grid.extent[0]
            e2 = self.model._grid.regular_grid.extent[3] - self.model._grid.regular_grid.extent[2]
            d1 = self.model._grid.regular_grid.extent[0]
            d2 = self.model._grid.regular_grid.extent[2]
            if len(list(centroids.items())[0][1]) == 2:
                c1, c2 = (0, 1)
            r1 = res[0]
            r2 = res[1]

        nkw = {
            "marker": "o",
            "color": "black",
            "markersize": 20,
            "alpha": 0.75
        }
        if node_kwargs is not None:
            nkw.update(node_kwargs)

        tkw = {
            "color": "white",
            "size": 10,
            "ha": "center",
            "va": "center",
            "weight": "ultralight",
            "family": "monospace"
        }
        if label_kwargs is not None:
            tkw.update(label_kwargs)

        lkw = {
            "linewidth": 0.75,
            "color": "black"
        }
        if edge_kwargs is not None:
            lkw.update(edge_kwargs)

        for edge in G.edges():
            a, b = edge
            # plot edges
            ax.plot(np.array([centroids[a][c1], centroids[b][c1]]) * e1 / r1 + d1,
                    np.array([centroids[a][c2], centroids[b][c2]]) * e2 / r2 + d2, **lkw)

            for node in G.nodes():
                ax.plot(centroids[node][c1] * e1 / r1 + d1, centroids[node][c2] * e2 / r2 + d2,
                        marker="o", color="black", markersize=10, alpha=0.75)
                ax.text(centroids[node][c1] * e1 / r1 + d1,
                        centroids[node][c2] * e2 / r2 + d2, str(node), **tkw)

    def plot_gradient(self, scalar_field, gx, gy, gz, cell_number, quiver_stepsize=5,
                      # maybe call r sth. like "stepsize"?
                      direction="y", plot_scalar=True, *args, **kwargs):  # include plot data?
        """
            Plot the gradient of the scalar field in a given direction.

            Args:
                geo_data (gempy.DataManagement.InputData): Input data of the model
                scalar_field(numpy.array): scalar field to plot with the gradient
                gx(numpy.array): gradient in x-direction
                gy(numpy.array): gradient in y-direction
                gz(numpy.array): gradient in z-direction
                cell_number(int): position of the array to plot
                quiver_stepsize(int): step size between arrows to indicate gradient
                direction(str): xyz. Caartesian direction to be plotted
                plot_scalar(bool): boolean to plot scalar field
                **kwargs: plt.contour kwargs

            Returns:
                None
        """

        raise NotImplementedError


def _scale_fig_size(figsize, textsize, rows=1, cols=1):
    """Scale figure properties according to rows and cols.

    Parameters
    ----------
    figsize : float or None
        Size of figure in inches
    textsize : float or None
        fontsize
    rows : int
        Number of rows
    cols : int
        Number of columns

    Returns
    -------
    figsize : float or None
        Size of figure in inches
    ax_labelsize : int
        fontsize for axes label
    titlesize : int
        fontsize for title
    xt_labelsize : int
        fontsize for axes ticks
    linewidth : int
        linewidth
    markersize : int
        markersize
    """
    params = mpl.rcParams
    rc_width, rc_height = tuple(params["figure.figsize"])
    rc_ax_labelsize = params["axes.labelsize"]
    rc_titlesize = params["axes.titlesize"]
    rc_xt_labelsize = params["xtick.labelsize"]
    rc_linewidth = params["lines.linewidth"]
    rc_markersize = params["lines.markersize"]
    if isinstance(rc_ax_labelsize, str):
        rc_ax_labelsize = 15
    if isinstance(rc_titlesize, str):
        rc_titlesize = 16
    if isinstance(rc_xt_labelsize, str):
        rc_xt_labelsize = 14

    if figsize is None:
        width, height = rc_width, rc_height
        sff = 1 if (rows == cols == 1) else 1.2
        width = width * cols * sff
        height = height * rows * sff
    else:
        width, height = figsize

    if textsize is not None:
        scale_factor = textsize / rc_xt_labelsize
    elif rows == cols == 1:
        scale_factor = ((width * height) / (rc_width * rc_height)) ** 0.5
    else:
        scale_factor = 1

    ax_labelsize = rc_ax_labelsize * scale_factor
    titlesize = rc_titlesize * scale_factor
    xt_labelsize = rc_xt_labelsize * scale_factor
    linewidth = rc_linewidth * scale_factor
    markersize = rc_markersize * scale_factor

    return (width, height), ax_labelsize, titlesize, xt_labelsize, linewidth, markersize



import numpy as np

from scipy.stats import entropy
import warnings


def compute_prob(blocks):
    warnings.warn("This function is Deprecated, please use the probability function instead")
    return probability(blocks)


def probability(blocks: np.ndarray) -> np.ndarray:
    """
    compute the probabilities for the unique values in the array given
    :param blocks: numpy array of different computed model results
    :return:
    """
    blocks = np.round(blocks)
    return np.mean([blocks == _ for _ in np.unique(blocks)], axis=1)


def calculate_ie_masked(prob):
    warnings.warn("This function is Deprecated, please use the information_entropy function instead")
    return information_entropy(prob)


def information_entropy(probabilities: np.ndarray, base=2) -> np.ndarray:
    """
    Calculate the Information Entropy of the provided probabilities array

    Eq 2. from doi.org/10.1016/j.tecto.2011.05.001

    but using scipy implementation for speed and simplicity

    note the paper assumes base two everywhere

    :param base: base for entropy, defaults to 2 as per paper
    :param probabilities: probabilities array, first axis is for each possible layers (M, n cells)
    :return: entropy of the probability array, flattens the first dimension
    """
    return entropy(probabilities, base=base)


def fuzziness(probabilities: np.ndarray) -> float:
    """
    Return the fuzziness of the probability array

    Eq 3. from doi.org/10.1016/j.tecto.2011.05.001
    :param probabilities: probabilities array
    :return: float of fuzziness
    """
    p = probabilities
    fuzz = -np.mean(np.nan_to_num(p * np.log(p) + (1 - p) * np.log(1 - p)))
    return fuzz


def total_model_entropy(ie: np.ndarray) -> float:
    """
    Return the Total Model Information Entropy (the mean of the array)
    Eq 4. from doi.org/10.1016/j.tecto.2011.05.001
    :param ie: information entropy array
    :return:
    """
    return float(np.mean(ie))



import numpy as np
import pandas as pn

from .surfaces import Surfaces


class Structure(object):
    """
    The structure_data class analyses the different lengths of subset in the interface and orientations categories_df
    to pass them to the aesara function.

    Attributes:
        surface_points (:class:`SurfacePoints`): [s0]
        orientations (:class:`Orientations`): [s1]
        surfaces (:class:`gempy.Surfaces`): [s2]
        faults (:class:`Faults`): [s3]
        df (:class:`pn.DataFrame`):
            * len surfaces surface_points (list): length of each surface/fault in surface_points
            * len series surface_points (list) : length of each series in surface_points
            * len series orientations (list) : length of each series in orientations
            * number surfaces per series (list): number of surfaces per series
            * ...
    Args:
        surface_points (:class:`SurfacePoints`): [s0]
        orientations (:class:`Orientations`): [s1]
        surfaces (:class:`gempy.Surfaces`): [s2]
        faults (:class:`Faults`): [s3]

    """

    def __init__(self, surface_points, orientations, surfaces: Surfaces, faults):
        self.surface_points = surface_points
        self.orientations = orientations
        self.surfaces = surfaces
        self.faults = faults

        df_ = pn.DataFrame(np.array(['False', 'False', -1, -1, -1, -1, -1, -1, -1], ).reshape(1, -1),
                           index=['values'],
                           columns=['isLith', 'isFault',
                                    'number faults', 'number surfaces', 'number series',
                                    'number surfaces per series',
                                    'len surfaces surface_points', 'len series surface_points',
                                    'len series orientations'])

        self.df = df_.astype({'isLith': bool, 'isFault': bool, 'number faults': int,
                              'number surfaces': int, 'number series': int})

        self.update_structure_from_input()

    def __repr__(self):
        return self.df.T.to_string()

    def _repr_html_(self):
        return self.df.T.to_html()

    def update_structure_from_input(self):
        """
        Update all fields dependent on the linked Data objects.

        Returns:
            bool: True
        """
        self.set_length_surfaces_i()
        self.set_series_and_length_series_i()
        self.set_length_series_o()
        self.set_number_of_surfaces_per_series()
        self.set_number_of_faults()
        self.set_number_of_surfaces()
        self.set_is_lith_is_fault()
        return True

    def set_length_surfaces_i(self):
        """
        Set the length of each **surface** on `SurfacePoints` i.e. how many data points are related to each surface

        Returns:
            :class:`pn.DataFrame`: df where Structural data is stored

        """
        # ==================
        # Extracting lengths
        # ==================
        # Array containing the size of every surface. SurfacePoints
        lssp = self.surface_points.df.groupby('id')['order_series'].count().values
        lssp_nonzero = lssp[np.nonzero(lssp)]

        self.df.at['values', 'len surfaces surface_points'] = lssp_nonzero

        return self.df

    def set_series_and_length_series_i(self):
        """
        Set the length of each **series** on `SurfacePoints` i.e. how many data points are related to each series. Also
        sets the number of series itself.

        Returns:
            :class:`pn.DataFrame`: df where Structural data is stored

        """
        len_series = self.surfaces.series.df.shape[0]

        # Array containing the size of every series. SurfacePoints.
        points_count = self.surface_points.df['order_series'].value_counts(sort=False)
        len_series_i = np.zeros(len_series, dtype=int)
        len_series_i[points_count.index.astype('int') - 1] = points_count.values

        if len_series_i.shape[0] == 0:
            len_series_i = np.insert(len_series_i, 0, 0)

        self.df.at['values', 'len series surface_points'] = len_series_i
        self.df['number series'] = len(len_series_i)
        return self.df

    def set_length_series_o(self):
        """
        Set the length of each **series** on `Orientations` i.e. how many orientations are related to each series.

        Returns:
            :class:`pn.DataFrame`: df where the Structural data is stored

        """
        # Array containing the size of every series. orientations.

        len_series_o = np.zeros(self.surfaces.series.df.shape[0], dtype=int)
        ori_count = self.orientations.df['order_series'].value_counts(sort=False)
        len_series_o[ori_count.index.astype('int') - 1] = ori_count.values

        self.df.at['values', 'len series orientations'] = len_series_o

        return self.df

    def set_number_of_surfaces_per_series(self):
        """
        Set number of surfaces for each series

        Returns:
            :class:`pn.DataFrame`: df where the Structural data is stored

        """
        len_sps = np.zeros(self.surfaces.series.df.shape[0], dtype=int)
        surf_count = self.surface_points.df.groupby('order_series'). \
            surface.nunique()

        len_sps[surf_count.index.astype('int') - 1] = surf_count.values

        self.df.at['values', 'number surfaces per series'] = len_sps
        return self.df

    def set_number_of_faults(self):
        """
        Set number of faults series. This method in gempy v2 is simply informative

        Returns:
            :class:`pn.DataFrame`: df where the Structural data is stored

        """
        # Number of faults existing in the surface_points df
        self.df.at['values', 'number faults'] = self.faults.df['isFault'].sum()
        return self.df

    def set_number_of_surfaces(self):
        """
        Set the number of total surfaces

        Returns:
            :class:`pn.DataFrame`: df where the Structural data is stored

        """
        # Number of surfaces existing in the surface_points df
        self.df.at['values', 'number surfaces'] = self.surface_points.df['surface'].nunique()

        return self.df

    def set_is_lith_is_fault(self):
        """
        Check if there are lithologies in the data and/or df. This method in gempy v2 is simply informative

        Returns:
            :class:`pn.DataFrame`: df where Structural data is stored
        """
        self.df['isLith'] = True if self.df.loc['values', 'number series'] >= self.df.loc['values', 'number faults'] \
            else False
        self.df['isFault'] = True if self.df.loc['values', 'number faults'] > 0 else False

        return self.df



import os

import pytest

import gempy_legacy
import gempy_legacy as gp
import pooch

from gempy_legacy.core.solution import xsolution_imported


def test_load_model(recompute=False):
    """Load model from disk"""
    cwd = os.path.dirname(__file__)
    data_path = cwd + '/../../examples/'

    if recompute:
        geo_model = gp.load_model(r'Tutorial_ch1-8_Onlap_relations',
                                  path=data_path + 'data/gempy_models/Tutorial_ch1'
                                                   '-8_Onlap_relations',
                                  recompile=True)
        gp.compute_model(geo_model)
        gp.plot_3d(geo_model, image=True)

    else:
        geo_model = gp.load_model(r'Tutorial_ch1-8_Onlap_relations',
                                  path=data_path + 'data/gempy_models/Tutorial_ch1'
                                                   '-8_Onlap_relations',
                                  recompile=False)


def test_save_model(one_fault_model_no_interp, tmpdir):
    """Save a model in a zip file with the default name and path"""
    gp.save_model(one_fault_model_no_interp, path=tmpdir)


# skip test is subsurface is not installed
@pytest.mark.skipif(not xsolution_imported, reason="Subsurface not installed")
def test_save_model_solution(one_fault_model_topo_solution, tmpdir):
    """Save a model in a zip file with the default name and path"""
    gp.save_model(
        model=one_fault_model_topo_solution,
        path=tmpdir,
        solution=True
    )
    print('Done')


def test_load_model_compressed():
    geo_model = gp.load_model(name="one_fault_model")


def test_load_model_compressed_remote():
    model_file = pooch.retrieve(url="https://github.com/cgre-aachen/gempy_data/raw/aesara_data/"
                                    "data/gempy_models/viz_3d.zip",
                                known_hash=None)

    geo_model = gp.load_model(name='viz_3d', path=model_file)


def test_load_model_compressed_remote_fail():
    with pytest.raises(Exception):
        model_file = pooch.retrieve(url="https://nowhere.zip",
                                    known_hash=None)

        geo_model = gp.load_model(name='error', path=model_file)


def test_load_model_compressed_remote2():
    model_file = pooch.retrieve(url="https://github.com/cgre-aachen/gempy_data/raw/aesara_data/"
                                    "data/gempy_models/Onlap_relations.zip",
                                known_hash=None)

    geo_model = gp.load_model(name='Onlap_relations', path=model_file, recompile=True)
    gp.compute_model(geo_model)
    gp.plot_3d(geo_model, image=True)


def test_pooch():
    goodboy = pooch.create(
        # Use the default cache folder for the OS
        path=pooch.os_cache("plumbus"),
        # The remote data is on Github
        base_url="https://raw.githubusercontent.com/cgre-aachen/gempy_data/master/data"
                 "/gempy_models/Tutorial_ch1-8_Onlap_relations/",
        # If this is a development version, get the data from the master branch
        version_dev="master",
        # We'll load it from a file later
        registry={
            "Tutorial_ch1-8_Onlap_relations_faults.csv": "19uheidhlkjdwhoiwuhc0uhcwljchw9ochwochw89dcgw9dcgwc"
        },
    )
    print(goodboy)



"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.
"""

import numpy as np
import pandas as pn
import matplotlib.pyplot as plt
#from gempy.plot.colors import *
#import matplotlib.cm as cm
#from gempy.plot.colors import color_lot, cmap, norm
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from gempy_legacy import Series


#import matplotlib.colors as mcolors

def set_anchor_points(series_object, surface_object):
    """
    Compute the location of each series and surface depending on the number of those

    Args:
        geo_model (gempy.data_management.InputData):

    Returns:
        list:
        - DataFrame: location of the series

        - DataFrame: location of the surfaces

        - float: thickness of the series

        - list, floats:

    """
    # Surfaces per serie
    series_names = series_object.df.index
    # Get number of series
    n_series = len(series_object.df.index)

    # Make anchor points for each serie
    anch_series_pos_aux = np.linspace(10, 0, n_series, endpoint=True)
    anch_series_pos = pn.DataFrame(anch_series_pos_aux.reshape(1, -1),
                                   columns=series_names)

    # Thicknes of series. We just make sure we have white space in between
    thick_series = 11.5 / n_series

    # Setting surfaces anchor
    anch_surfaces_pos = pn.DataFrame()
    thick_surfaces = []
    for series in series_names:
        try:
            surfaces = surface_object.set_index('series').loc[series, 'surface']
            #if type(surfaces) is not str or type(surfaces) is not np.str_:
            try:
                surfaces = surfaces.values
            except AttributeError:
                pass
        except KeyError:
            surfaces = np.empty(0, dtype='object')

        surfaces = np.insert(surfaces, 0, '0_aux' + series)
        surfaces = np.append(surfaces, '1_aux' + series)
        anch_for_df = pn.DataFrame(
            np.linspace(anch_series_pos[series][0] + thick_series / 2,
                        anch_series_pos[series][0] - thick_series / 2,
                        surfaces.shape[0],
                        endpoint=True).reshape(1, -1),
            columns=surfaces)

        anch_surfaces_pos = pn.concat([anch_surfaces_pos, anch_for_df],
                                        axis=1)
        thick_surfaces = np.append(thick_surfaces,
                                     (np.tile((thick_series - 2) / surfaces.shape[0], surfaces.shape[0])))

    return anch_series_pos, anch_surfaces_pos, thick_series, thick_surfaces

plt.style.use(['seaborn-white', 'seaborn-talk'])


class StratigraphicPile(object):
    """
    Class to create the interactive stratigraphic pile
    """
    def __init__(self, series_class, surface_class):

        # Set the values of matplotlib
        fig = plt.figure()
        ax = fig.add_subplot(111)
        ax.set_xlim(0, 7)
        ax.get_yaxis().set_visible(False)
        ax.get_xaxis().set_visible(False)
        ax.axis('off')

        # Compute the anchor values for the number of series. This is a DataFrame
        self.anch_series, self.anch_surfaces, self.thick_series, self.thick_surfaces = set_anchor_points(series_class,
                                                                                                             surface_class)

        # We create the list that contains rectangles that represent our series ang are global
        global series_rect
        series_rect = {}

        # Define the initial value of each rectangle
        pos_anch = np.squeeze(self.anch_series.values)
        rects = ax.barh(pos_anch, np.ones_like(pos_anch)*2, self.thick_series, )

        stratcols = ['#400505', '#7D050E', '#646363', '#878786', '#AFAEAE', '#D9D9D8', '#EAEA3B', '#DDAF19']
        # We connect each rectangle
        for e, series in enumerate(series_class.df.index):
            # TODO Alex set the colors of the series accordingly

            rects[e].set_color(stratcols[e])
            rects[e].set_label(series)
            dr = DraggableRectangle(rects[e], series_class, surface_class)
            dr.connect()
            dr.rect.f = None
            dr.rect.s = series

            series_rect[series] = dr

        global surface_rect
        surface_rect = {}

        # Define the initial value of each rectangle
        pos_anch = np.squeeze(self.anch_surfaces.values)
        rects = ax.barh(pos_anch, np.ones_like(pos_anch)*2, .5, left=3.)

        color_lot = dict(zip(surface_class['surface'], surface_class['color']))
        #color_lot = surface_class.colors.colordict
        # We connect each rectangle
        for e, surface in enumerate(self.anch_surfaces.columns):
            if 'aux' in surface:
                rects[e].set_alpha(.1)
                rects[e].set_color('gray')
            else:
                try:
                    rects[e].set_color(color_lot[surface])
                    rects[e].set_label(surface)
                except KeyError:
                    pass
            dr = DraggableRectangle(rects[e], series_class, surface_class)
            dr.connect()
            dr.rect.f = surface
            dr.rect.s = None

            surface_rect[surface] = dr
        plt.legend(bbox_to_anchor=(1.1, 1.05))
        plt.ion()
        ax.text(1, self.anch_series.max().values.max() + self.thick_series/2 + 2, r'Series', fontsize=15,
                fontweight='bold', bbox={'facecolor':'gray', 'alpha':0.5, 'pad':10}, horizontalalignment='center')
        ax.text(4, self.anch_series.max().values.max() + self.thick_series/2 + 2, r'Faults/Surfaces', fontsize=15,
                fontweight='bold', bbox={'facecolor':'gray', 'alpha':0.5, 'pad':10}, horizontalalignment='center')

        self.figure = plt.gcf()
        plt.close(self.figure)


class DraggableRectangle:
    def __init__(self, rect, series, surface_df):
        # The idea of passing geodata is to update the dataframes in place
        self.series = series
        self.surface = surface_df
        self.rect = rect

        # We add the name of the series as attribute of the rectangle
        self.press = None

        # We initalize the placement of the anchors
        self.anch_series, self.anch_surfaces, self.thick_series, self.thick_surfaces = set_anchor_points(self.series,
                                                                                                             surface_df)

    def connect(self):
        'connect to all the events we need'
        self.cidpress = self.rect.figure.canvas.mpl_connect(
            'button_press_event', self.on_press)
        self.cidrelease = self.rect.figure.canvas.mpl_connect(
            'button_release_event', self.on_release)
        self.cidmotion = self.rect.figure.canvas.mpl_connect(
            'motion_notify_event', self.on_motion)

    def on_press(self, event):
        'on button press we will see if the mouse is over us and store some data'
        if event.inaxes != self.rect.axes: return

        contains, attrd = self.rect.contains(event)

        if not contains: return

        x0, y0 = self.rect.xy

        # We detect the series that has been touched.
        self.selected_rectangle_s = self.rect.s
        self.selected_rectangle_f = self.rect.f

        # We pass all the important attributes through this property
        self.press = x0, y0, event.xdata, event.ydata, self.selected_rectangle_s, self.selected_rectangle_f

    def on_motion(self, event):
        'on motion we will move the rect if the mouse is over us'
        if self.press is None: return
        if event.inaxes != self.rect.axes: return
        x0, y0, xpress, ypress = self.press[:-2]

        # We forbid movement in x
        dy = event.ydata - ypress
        self.rect.set_y(y0 + dy)
        self.rect.figure.canvas.draw()

    def on_release(self, event):
        'on release we reset the press data'

        # We extract the rectangle that was clicked
        if self.press is None: return
        selected_rectangle_s = self.press[-2]
        selected_rectangle_f = self.press[-1]
        self.press = None
        self.rect.figure.canvas.draw()

        if self.selected_rectangle_s is not None:
            # Make a copy of the position where the selected rectangle was
            selected_arch_position_s = np.copy(self.anch_series[selected_rectangle_s].values)

            # Compute the position of the closes anchor point and the argument of the rectangle which was there
            dropping_arch_position, dropping_arg = self.compute_new_arch_series()

            # Here we change the selected rectangle to the position to the dropping position
            self.anch_series[selected_rectangle_s] = dropping_arch_position

            # Here we change the rectangle that was on the dropping position to the original position of the selected
            # rectangle
            self.anch_series.iloc[0, dropping_arg] = selected_arch_position_s

            self.update_data_frame()
            self.anch_series, self.anch_surfaces, self.thick_series, self.thick_surfaces = set_anchor_points(
                self.series, self.surface)

            # We update the visualization of all the rectangles
            for series_name in self.anch_series.columns:
                series_rect[series_name].rect.set_y(self.anch_series[series_rect[series_name].rect.s].values-self.thick_series/2)
                series_rect[series_name].rect.set_animated(False)
                series_rect[series_name].rect.background = None

                # redraw the full figure
                series_rect[series_name].rect.figure.canvas.draw()

        if self.selected_rectangle_f is not None:

            # Make a copy of the position where the selected rectangle was
            selected_arch_position_f = np.copy(self.anch_surfaces[selected_rectangle_f].values)
            # selected_arch_position_f = np.copy(self.anch_surfaces[selected_rectangle].values)

            # Compute the position of the closes anchor point and the argument of the rectangle which was there
            dropping_arch_position, dropping_arg = self.compute_new_arch_surface()

            # Here we change the selected rectangle to the position to the dropping position
            self.anch_surfaces[selected_rectangle_f] = dropping_arch_position

            # Here we change the rectangle that was on the dropping position to the original position of the selected
            # rectangle
            self.anch_surfaces.iloc[0, dropping_arg] = selected_arch_position_f

            self.update_data_frame()
            self.anch_series, self.anch_surfaces, self.thick_series, self.thick_surfaces = set_anchor_points(
                self.series, self.surface)

        # We update the visualization of all the rectangles
        for surfaces_name in self.anch_surfaces.columns:
            surface_rect[surfaces_name].anch_surfaces = self.anch_surfaces
            new_pos = self.anch_surfaces[surface_rect[surfaces_name].rect.f].values
            surface_rect[surfaces_name].rect.set_y(new_pos - 0.5/2)
            surface_rect[surfaces_name].rect.set_animated(False)
            surface_rect[surfaces_name].rect.background = None

            # redraw the full figure
            surface_rect[surfaces_name].rect.figure.canvas.draw()

        self.selected_rectangle_s = None
        self.selected_rectangle_r = None

    def compute_new_arch_series(self):

        dist = np.abs(self.anch_series.values - self.rect.get_y())
        arg_min = np.argmin(dist)
        new_arch = self.anch_series.iloc[0, arg_min]
        return new_arch, arg_min

    def compute_new_arch_surface(self):

        dist = np.abs(self.anch_surfaces.values - self.rect.get_y())
        arg_min = np.argmin(dist)
        new_arch = self.anch_surfaces.iloc[0, arg_min]
        return new_arch, arg_min

    def update_data_frame(self):

        order_series = self.anch_series.sort_values(by=0, axis=1, ascending=False).columns.values
        order_surfaces = self.anch_surfaces.sort_values(by=0, axis=1, ascending=False)

        # drop aux
        aux_columns = ['aux' in i for i in order_surfaces.columns]
        order_surfaces.drop(order_surfaces.columns[aux_columns], axis=1, inplace=True)

        series_dict = {}
        # divide surfaces to their series
        for name, value in self.anch_series.iteritems():
             cond = ((order_surfaces <= value.get_values()+self.thick_series/2) & \
                    (order_surfaces >= value.get_values()-self.thick_series/2))
             format_in_series = order_surfaces.columns[cond.values[0, :]]
             # Passing from array to tuple what a pain
             series_dict[name] = format_in_series.values

        self.surface.map_series(series_dict)
        #self.series.set_surface_number(order_surfaces.columns.values)
        #self.geo_model.order_table()

    def disconnect(self):
        'disconnect all the stored connection ids'
        self.rect.figure.canvas.mpl_disconnect(self.cidpress)
        self.rect.figure.canvas.mpl_disconnect(self.cidrelease)
        self.rect.figure.canvas.mpl_disconnect(self.cidmotion)






# coding: utf-8

# # 3 - 3D Modeling with GemPy

# In[1]:

from matplotlib import use
use("Agg")

import sys
import numpy as np
# These two lines are necessary only if gempy is not installed
sys.path.append("../../gempy/")
sys.path.append("../../../remote-geomod/rgeomod")
sys.path.append("../gempy/")
sys.path.append("/home/miguel/PycharmProjects/remote-geomod/")
# Importing gempy
import gempy_legacy as gp
import pytest

#import gdal
#import rgeomod
import warnings
warnings.filterwarnings('ignore')
import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
import os

gdal = pytest.importorskip("gdal")
rgeomod = pytest.importorskip("rgeomod")
input_path = os.path.dirname(__file__)+'/../input_data'


def TOUPDATE_rgeomod_integration(aesara_f):
    geo_data=gp.create_data(extent=[612000, 622000, 2472000, 2480000, -1000, 1000],
                            resolution=[50, 50, 50],
                            path_f=input_path+"/gempy_foliations.csv",
                            path_i=input_path+"/gempy_interfaces.csv")



    formation_order = ["Unit4", "Unit3", "Unit2", "Unit1"]



    gp.set_series(geo_data, {"Default series": formation_order},
                 order_formations = formation_order, verbose=1)



    gp.plot_data(geo_data, direction="z")


    #interp_data = gp.InterpolatorData(geo_model, compile_aesara=True)
    interp_data = aesara_f
    interp_data.update_interpolator(geo_data)

    lith_block, fault_block = gp.compute_model(interp_data)
    print("3-D geological model calculated.")


    gp.plot_section(geo_data, lith_block[0], 25, direction='y', plot_data=False)
    #plt.savefig("../data/cross_section_NS_25.pdf", bbox_inches="tight")

    gp.plot_section(geo_data, lith_block[0], 25, direction='x', plot_data=False)
    #plt.savefig("../data/cross_section_EW_25.pdf", bbox_inches="tight")

    vertices, simplices = gp.get_surfaces(interp_data, potential_lith=lith_block[1], step_size=2)

    fig = plt.figure(figsize=(13,10))
    ax = fig.add_subplot(111, projection='3d')
    cs = ["lightblue", "pink", "lightgreen", "orange"]
    for i in range(4):
        surf = ax.plot_trisurf(vertices[i][:,0], vertices[i][:,1], vertices[i][:,2],
                               color=cs[i], linewidth=0, alpha=0.65, shade=False)
    #plt.savefig("../data/surfaces_3D.pdf", bbox_inches="tight")

    # try:
    #     gp.plot_surfaces_3D(geo_model, vertices, simplices)
    # except NameError:
    #     print("3-D visualization library vtk not installed.")

    # load the digital elevation model
    geotiff_filepath = input_path+"/dome_sub_sub_utm.tif"
    raster = gdal.Open(geotiff_filepath)
    dtm = raster.ReadAsArray()
    dtmp = plt.imshow(dtm, origin='upper', cmap="viridis");
    plt.title("Digital elevation model");
    plt.colorbar(dtmp, label="Elevation [m]");
    plt.savefig(input_path+"temp/DTM.pdf")

    # To be able to use gempy plot functionality we need to create a dummy geo_model object with the
    # resoluion we want. In this case resolution=[339, 271, 1]
    import copy
    geo_data_dummy = copy.deepcopy(geo_data)
    geo_data_dummy.resolution = [339, 271, 1]



    # convert the dtm to a gempy-suitable raveled grid
    points = rgeomod.convert_dtm_to_gempy_grid(raster, dtm)


    # Now we can use the function `compute_model_at` to get the lithology values at a specific location:

    # In[17]:


    # interp_data_geomap = gp.InterpolatorInput(geo_model, dtype="float64")
    lith_block, fault_block = gp.compute_model_at(points, interp_data)


    # <div class="alert alert-info">
    # **Your task:** Create a visual representation of the geological map in a 2-D plot (note: result is also again saved to the `../data`-folder):
    # </div>
    #
    # And here **the geological map**:

    # In[18]:


    gp.plot_section(geo_data_dummy, lith_block[0], 0, direction='z', plot_data=False)
    plt.title("Geological map");
    #plt.savefig("../geological_map.pdf")


    # ### Export the map for visualization in GoogleEarth

    # <div class="alert alert-info">
    # **Your task:** Execute the following code to export a GeoTiff of the generated geological map, as well as `kml`-files with your picked points inside the data folder. Open these files in GoogleEarth and inspect the generated map:
    # </div>
    #
    #
    # <div class="alert alert-warning">
    # **Note (1)**: Use the normal `File -> Open..` dialog in GoogleEarth to open the data - no need to use the `Import` method, as the GeoTiff contains the correct coordinates in the file.
    # </div>
    #
    #
    # <div class="alert alert-warning">
    # **Note (2)**: For a better interpretation of the generated map, use the transparency feature (directly after opening the map, or using `right-click -> Get Info` on the file).
    # </div>

    # In[19]:


    geo_map = lith_block[0].copy().reshape((339,271))
    geo_map = geo_map.astype('int16')  # change to int for later use


    # In[20]:


    rgeomod.export_geotiff(input_path+"temp/geomap.tif", geo_map, gp.plotting.colors.cmap, geotiff_filepath)


    # Export the interface data points:

    # In[21]:


    t = input_path+"/templates/ge_template_raw_interf.xml"
    pt = input_path+"/templates/ge_placemark_template_interf.xml"
    rgeomod.gempy_export_points_to_kml(input_path, geo_data, pt, t, gp.plotting.colors.cmap)


    # Export the foliation data:



    t = input_path+"/templates/ge_template_raw_fol.xml"
    pt = input_path+"/templates/ge_placemark_template_fol.xml"
    rgeomod.gempy_export_fol_to_kml(input_path+"temp/dips.kml", geo_data, pt, t)



import pytest
import sys, os

from gempy_legacy.addons.gempy_to_rexfile import GemPyToRex

sys.path.append("../../..")
import gempy_legacy

request = pytest.importorskip("requests")
from gempy_legacy.addons import gempy_to_rexfile as gtr
from gempy_legacy.addons import rex_api

pyqrcode = pytest.importorskip("pyqrcode")
input_path = os.path.dirname(__file__) + '/../input_data'


@pytest.mark.skipif("TRAVIS" in os.environ and os.environ["TRAVIS"] == "true",
                    reason="Skipping this test on Travis CI.")
class TestGemPyToREX:
    @pytest.fixture(scope='module')
    def geo_model(self, model_horizontal_two_layers):
        geo_data = model_horizontal_two_layers

        # Compute model
        sol = gempy_legacy.compute_model(geo_data, compute_mesh_options={'rescale': True})

        return geo_data

    def test_write_header(self):
        header_bytes = gtr.write_file_header_block(3, 1)
        if False:
            gtr.write_file(header_bytes, 'DEP/rexfiles/header_test')

    def test_write_mesh(self, geo_model):
        mesh_header_size = 128
        file_header_size = 86
        ver, tri = geo_model.solutions.vertices[0], geo_model.solutions.edges[0]

        ver_ravel, tri_ravel, n_vtx_coord, n_triangles = gtr.mesh_preprocess(ver, tri)
        data_block_size_no_header = (n_vtx_coord + n_triangles) * 4 + mesh_header_size

        # Write header
        header_bytes = gtr.write_file_header_block(n_data_blocks=1,
                                                   size_data_blocks=data_block_size_no_header + gtr.rexDataBlockHeaderSize,
                                                   start_data=file_header_size)

        # Write data block
        data_bytes = gtr.write_data_block_header(size_data=data_block_size_no_header,
                                                 data_id=1, data_type=3, version_data=1)

        # Write mesh block
        mesh_header_bytes = gtr.write_mesh_header(n_vtx_coord / 3, n_triangles / 3,
                                                  start_vtx_coord=gtr.mesh_header_size,
                                                  start_nor_coord=gtr.mesh_header_size + n_vtx_coord * 4,
                                                  start_tex_coord=gtr.mesh_header_size + n_vtx_coord * 4,
                                                  start_vtx_colors=gtr.mesh_header_size + n_vtx_coord * 4,
                                                  start_triangles=gtr.mesh_header_size + n_vtx_coord * 4,
                                                  name='test_a')

        mesh_block_bytes = gtr.write_mesh_coordinates(ver_ravel, tri_ravel)

        all_bytes = header_bytes + data_bytes + mesh_header_bytes + mesh_block_bytes

        if False:
            gtr.write_file(all_bytes, 'DEP/rexfiles/one_mesh_test')

    def TEST_rex_cloud_api(self):
        import datetime
        timestamp = datetime.datetime.now()
        project_name = str(timestamp)
        rex_api.RexAPI(project_name)

    def test_geo_model_to_rex(self, geo_model):
        rex_bytes = gtr.geomodel_to_rex(geo_model)
        print(len(rex_bytes['A']))
        assert len(rex_bytes['A']) == 235706

    def test_rex_bytes_to_file(self, geo_model):
        rex_bytes = gtr.geomodel_to_rex(geo_model)
        gtr.write_rex(rex_bytes, path=os.path.dirname(__file__) + '/rexfiles/gtr_test')

    def test_rex_bytes_to_file_except(self):
        model = gempy_legacy.create_data(extent=[0, 10, 0, 10, 0, 10],
                                         resolution=[50, 50, 50])
        model.set_default_surfaces()
        model._surfaces.df['isActive'] = False
        with pytest.raises(RuntimeError):
            rex_bytes = gtr.geomodel_to_rex(model)

    @pytest.mark.skip(reason="Needs token and secret. (@leguark they are in notion)")
    def test_plot_ar(self, geo_model):
        tag = gempy_legacy.plot.plot_ar(geo_model,
                                        api_token='foo', \
                                        project_name='Alesmodel',
                                        secret='bar')
        print(tag.display_tag(reverse=False))



import gempy_legacy as gp
import numpy as np


def test_colors_101_surfaces():
    """Tests if GemPy Colors class works with at least 101 surfaces."""
    geomodel = gp.create_model("ColorfulModel")
    for n in range(101):
        geomodel.add_surfaces(f"Surface {n}")
    assert np.all(geomodel.surfaces.df.color.values != np.nan)


"""
    This file is part of gempy.

    gempy is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    gempy is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with gempy.  If not, see <http://www.gnu.org/licenses/>.


@author: Alexander Schaaf
"""
from warnings import warn
import numpy as np
try:
    from skimage.measure import regionprops, label
except ImportError:
    warn("skimage package is not installed, which is required for geomodel complexity analysis.")


def get_nearestneighbor_block(lb):
    """
    Simplest 3D nearest neighbor comparison to (6-stamp) for lithology values.

    Args:
        lb (np.ndarray): lb[0].reshape(*geo_data.resolution).astype(int)

    Returns:
        (np.ndarray)
    """
    shp = lb.shape
    nn = np.zeros((shp[0] - 1, shp[0] - 1, shp[0] - 1))
    # x
    nn += np.abs(lb[1:, :-1, :-1] ^ lb[:-1, :-1, :-1])
    nn += np.abs(lb[:-1, :-1, :-1] ^ lb[1:, :-1, :-1])
    # y
    nn += np.abs(lb[:-1, 1:, :-1] ^ lb[:-1, :-1, :-1])
    nn += np.abs(lb[:-1, :-1, :-1] ^ lb[:-1, 1:, :-1])
    # z
    nn += np.abs(lb[:-1, :-1, 1:] ^ lb[:-1, :-1, :-1])
    nn += np.abs(lb[:-1, :-1, :-1] ^ lb[:-1, :-1, 1:])

    return nn


def get_geobody_volume(rprops, geo_data=None):
    """Compute voxel counts per unique integer body in given block model

    Args:
        rprops (list): List of regionprops object for each unique region of the model.
        (skimage.measure._regionprops._RegionProperties object)

    Returns:
        (dict): Dict with node labels as keys and geobody volume values.
    """
    if geo_data is None:
        return {rprop.label:rprop.area for rprop in rprops}
    else:
        return {rprop.label:rprop.area * np.product(geo_data.extent[1::2] / geo_data.resolution) for rprop in rprops}


def get_geobody_tops(rprops, geo_data=None):
    """Get the top vertical limit coordinate of geobodies (via bbox).

    Args:
        rprops (list): List of regionprops object for each unique region of the model.
        (skimage.measure._regionprops._RegionProperties object)
        geo_data (gempy.data_management.InputData):

    Returns:
        (dict): Dict with node labels as keys and geobody top coordinates as values.
    """

    if geo_data is None:
        return {rprop.label: rprop.bbox[5] for rprop in rprops}
    else:
        return {rprop.label: rprop.bbox[5] * geo_data.extent[5] / geo_data.resolution[2] for rprop in rprops}


def get_geobody_bots(rprops, geo_data=None):
    """Get the bottom vertical limit coordinate of geobodies (via bbox).

    Args:
        rprops (list): List of regionprops object for each unique region of the model.
        (skimage.measure._regionprops._RegionProperties object)
        geo_data (gempy.data_management.InputData):

    Returns:
        (dict): Dict with node labels as keys and geobody bottom coordinates as values.
    """
    if geo_data is None:
        return {rprop.label: rprop.bbox[2] for rprop in rprops}
    else:
        return {rprop.label: rprop.bbox[2] * geo_data.extent[5] / geo_data.resolution[2] for rprop in rprops}


def get_centroids(rprops):
    """Get node centroids in 2d and 3d as {node id (int): tuple(x,y,z)}."""
    centroids = {}
    for rp in rprops:
            centroids[rp.label] = rp.centroid
    return centroids


def get_unique_regions(lith_block, fault_block, n_faults, neighbors=8, noddy=False):
    """

    Args:
        lith_block (np.ndarray): Lithology block model
        fault_block (np.ndarray): Fault block model
        n_faults (int): Number of faults.
        neighbors (int, optional): Specifies the neighbor voxel connectivity taken into account for the topology
            analysis. Must be either 4 or 8 (default: 8)
        noddy (bool): If a noddy block is handed to the function, equalizes the results to be comparable with GemPy

    Returns:
        (np.ndarray): Model block with uniquely labeled regions.
    """
    lith_block = np.round(lith_block).astype(int)
    fault_block = np.round(fault_block).astype(int)

    # label the fault block for normalization (comparability of e.g. pynoddy and gempy models)
    fault_block = label(fault_block, neighbors=neighbors, background=9999)

    if noddy:
        # then this is a gempy model, numpy starts with 1
        lith_block[lith_block == 0] = int(np.max(lith_block) + 1)  # set the 0 to highest value + 1
        lith_block -= n_faults  # lower by n_faults to equal with pynoddy models
        # so the block starts at 1 and goes continuously to max

    ublock = (lith_block.max() + 1) * fault_block + lith_block
    labels_block, labels_n = label(ublock, neighbors=neighbors, return_num=True, background=9999)
    if 0 in np.unique(labels_block):
        labels_block += 1

    return labels_block


def calculate_probability_lithology(lith_blocks):
    """Blocks must be just the lith blocks!"""
    lith_id = np.unique(lith_blocks)
    # lith_count = np.zeros_like(lith_blocks[0:len(lith_id)])
    lith_count = np.zeros((len(np.unique(lith_blocks)), lith_blocks.shape[1]))
    for i, l_id in enumerate(lith_id):
        lith_count[i] = np.sum(lith_blocks == l_id, axis=0)
    lith_prob = lith_count / len(lith_blocks)
    return lith_prob


def calculate_information_entropy(lith_prob):
    """Calculates information entropy for the given probability array."""
    ie = np.zeros_like(lith_prob[0])
    for l in lith_prob:
        pm = np.ma.masked_equal(l, 0)  # mask where layer prob is 0
        ie -= (pm * np.ma.log2(pm)).filled(0)
    return ie


def calculate_information_entropy_total(ie, absolute=False):
    """Calculate total information entropy (float) from an information entropy array."""
    if absolute:
        return np.sum(ie)
    else:
        return np.sum(ie) / np.size(ie)


def calculate_ie(lbs):
    """Computes the per-voxel and total information entropy of given block models."""

    lith_prob = calculate_probability_lithology(lbs)
    ie = calculate_information_entropy(lith_prob)
    ie_total = calculate_information_entropy_total(ie)

    return ie, ie_total


import numpy as np
import matplotlib.pyplot as plt
import gempy_legacy as gp


def plot_sig(n_surface_0, n_surface_1, a, b, drift,
             l=50, Z_x=None, sf_max=None, sf_min=None, sf_at_scalar=None, relu=None, **kwargs):
    """
    Plot the sigmoid function used by gempy to discretize space

    Args:
        n_surface_0 (list, int): Activation value
        n_surface_1 (list, int): Deactivation value
        a (list, float): value of the scalar field to trigger the activation
        b (list, float): value of the scalar field to trigger the deactivation
        drift (list, float): drift of the function
        l (float): sigmoid slope
        Z_x (ndarray): scalar field vector
        sf_max (Optional[float]): Maximum scalar field of the model. This represent the boundaries
        sf_min (Optional[float]): Minimum scalar field of the model. This represent the boundaries
        sf_at_scalar (Optional[ndarray]): Values where the layers are found

    Keyword Args:
        colors: List with the colors for the layers

    Returns:

    """
    if 'colors' not in kwargs:
        colors = ['#015482','#9f0052','#ffbe00','#728f02','#443988','#ff3f20','#325916','#5DA629']

    if Z_x is None:
        Z_x = np.linspace(-3, 3, 2000)
    f_x_s = np.zeros_like(Z_x)
    relu_up = np.copy(Z_x)
    relu_up -= b[0]
    relu_up[relu_up < 0] = 0
    relu_up = relu_up * -.01

    # print(relu_up)
    relu_down = np.copy(Z_x)
    relu_down -= a[-1]
    relu_down[relu_down > 0] = 0
    relu_down = relu_down * -.01

    relu = relu_up + relu_down

    if len(n_surface_0) == 1:

        f_x = -n_surface_0 / (1 + np.exp(-l * (Z_x - a))) - \
              (n_surface_1 / (1 + np.exp(l * (Z_x - b)))) + drift
        f_x += relu

        plt.plot(f_x, Z_x)

    else:
        len_ = len(n_surface_0)
        fig = plt.figure(figsize=(7, 12))
        for e in range(len_):
            f_x = - n_surface_0[e] / (1 + np.exp(-l * (Z_x - a[e]))) - \
                  (n_surface_1[e] / (1 + np.exp(l * (Z_x - b[e])))) + drift[e]
            f_x_s += f_x + relu
            # fig.add_subplot(len_, 1, e+1)
            plt.plot(f_x, Z_x, '--', label='Layer ' + str(drift[e]))

        if sf_max is not None:
            plt.hlines(sf_max, 0, f_x_s.max(), label='Model Extent')
        if sf_min is not None:
            plt.hlines(sf_min, 0, f_x_s.max())
        if sf_at_scalar is not None:
            plt.hlines(sf_at_scalar, 0, f_x_s.max(), linewidth=3,
                       color=colors[:len(sf_at_scalar)], label='Scalar value interfaces')

        plt.plot(f_x_s, Z_x, linewidth=5, alpha=.7, label='Actual exporty')
        plt.ylabel('Scalar field')
        plt.xlabel('Lith block')
        #  plt.gca().invert_yaxis()

        plt.legend(bbox_to_anchor=(1.8, 1))

    return plt.gcf()






